from __future__ import annotations

import os
import sys
import shutil
import typing as t
import platform
import tempfile
import subprocess
from abc import ABCMeta
from enum import Enum, unique
from pathlib import Path
from functools import partial
from collections import defaultdict

import yaml
import jinja2
from fs import open_fs
from fs.copy import copy_fs, copy_file
from fs.tarfs import TarFS
from typing_extensions import Protocol

from starwhale.utils import (
    docker,
    console,
    load_yaml,
    in_container,
    validate_obj_name,
    make_dir_gitignore,
    get_downloadable_sw_version,
)
from starwhale.consts import (
    SupportOS,
    LATEST_TAG,
    SupportArch,
    PythonRunEnv,
    SW_IMAGE_FMT,
    CREATED_AT_KEY,
    DEFAULT_PROJECT,
    DefaultYAMLName,
    SW_AUTO_DIRNAME,
    DEFAULT_PAGE_IDX,
    SW_PYPI_PKG_NAME,
    DEFAULT_PAGE_SIZE,
    ENV_SW_IMAGE_REPO,
    DEFAULT_IMAGE_REPO,
    STANDALONE_INSTANCE,
    SW_DEV_DUMMY_VERSION,
    WHEEL_FILE_EXTENSION,
    DEFAULT_CONDA_CHANNEL,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.version import STARWHALE_VERSION
from starwhale.base.tag import StandaloneTag
from starwhale.utils.fs import (
    move_dir,
    empty_dir,
    ensure_dir,
    ensure_file,
    is_within_dir,
    get_path_created_time,
)
from starwhale.base.type import (
    BundleType,
    InstanceType,
    DependencyType,
    RuntimeArtifactType,
    RuntimeLockFileType,
)
from starwhale.base.cloud import CloudRequestMixed, CloudBundleModelMixin
from starwhale.base.mixin import ASDictMixin
from starwhale.utils.http import ignore_error
from starwhale.utils.venv import (
    is_venv,
    is_conda,
    venv_setup,
    conda_setup,
    conda_export,
    get_base_prefix,
    get_conda_pybin,
    conda_env_update,
    extract_venv_pkg,
    venv_install_req,
    conda_install_req,
    create_python_env,
    extract_conda_pkg,
    install_starwhale,
    get_python_version,
    package_python_env,
    activate_python_env,
    pip_freeze_by_pybin,
    guess_current_py_env,
    trunc_python_version,
    get_conda_prefix_path,
    check_valid_venv_prefix,
    get_user_python_version,
    check_valid_conda_prefix,
    get_python_version_by_bin,
    render_python_env_activate,
    get_user_runtime_python_bin,
    pip_compatible_dependencies_check,
)
from starwhale.base.bundle import BaseBundle, LocalStorageBundleMixin
from starwhale.utils.error import (
    FormatError,
    ExistedError,
    NotFoundError,
    NoSupportError,
    ConfigFormatError,
    MissingFieldError,
    ExclusiveArgsError,
    UnExpectedConfigFieldError,
)
from starwhale.utils.progress import run_with_progress_bar
from starwhale.base.bundle_copy import BundleCopy
from starwhale.base.uri.project import Project as ProjectURI
from starwhale.base.uri.resource import Resource, ResourceType

from .store import RuntimeStorage

RUNTIME_API_VERSION = "1.1"
_TEMPLATE_DIR = Path(__file__).parent / "template"

_t_mixed_str_list = t.Union[t.List[str], str]
_list: t.Callable[[_t_mixed_str_list], t.List[str]] = (
    lambda _x: _x if isinstance(_x, (list, tuple)) else [_x]
)

_SUPPORT_CUDA = ["11.3", "11.4", "11.5", "11.6", "11.7"]
_SUPPORT_CUDNN = {"8": {"support_cuda_versions": ["11.3", "11.4", "11.5", "11.6"]}}
_SUPPORT_PYTHON_VERSIONS = ["3.7", "3.8", "3.9", "3.10", "3.11"]


@unique
class RuntimeInfoFilter(Enum):
    basic = "basic"
    runtime_yaml = "runtime_yaml"
    manifest = "manifest"
    lock = "lock"
    all = "all"


class DockerEnv(ASDictMixin):
    def __init__(self, **kwargs: t.Any):
        self.image = kwargs.get("image", "")

    def __str__(self) -> str:
        if self.image:
            return f"image:{self.image}"
        return "empty"

    __repr__ = __str__


class Environment(ASDictMixin):
    def __init__(
        self,
        arch: _t_mixed_str_list = "",
        os: str = SupportOS.UBUNTU,
        python: str = "",
        cuda: str = "",
        cudnn: str = "",
        **kw: t.Any,
    ) -> None:
        self.arch = _list(arch)
        self.os = os.lower()

        if not python:
            python = get_python_version()
        self.python = trunc_python_version(str(python))
        self.cuda = str(cuda).strip()
        self.cudnn = str(cudnn).strip()
        self.docker = DockerEnv(**kw.get("docker", {}))

        self._do_validate()

    def _do_validate(self) -> None:
        if self.os not in (SupportOS.UBUNTU,):
            raise NoSupportError(f"environment.os {self.os}")

        if self.python not in _SUPPORT_PYTHON_VERSIONS:
            raise ConfigFormatError(
                f"only support Python[{_SUPPORT_PYTHON_VERSIONS}], set {self.python}"
            )

        if self.cuda and self.cuda not in _SUPPORT_CUDA:
            raise NoSupportError(
                f"cuda {self.cuda} no support[supported list: {_SUPPORT_CUDA}]"
            )

        if self.cudnn and self.cudnn not in _SUPPORT_CUDNN:
            raise NoSupportError(
                f"cudnn {self.cudnn} no support[supported list: {_SUPPORT_CUDNN}]"
            )

        if self.cudnn and not self.cuda:
            raise MissingFieldError("cuda is the precondition for cudnn")

        if (
            self.cuda
            and self.cudnn
            and self.cuda not in _SUPPORT_CUDNN[self.cudnn]["support_cuda_versions"]
        ):
            raise NoSupportError(f"cuda:{self.cuda} no support cudnn:{self.cudnn}")

    def __str__(self) -> str:
        return f"Starwhale Runtime Environment: {self.os}-{self.arch}-python:{self.python}-cuda:{self.cuda}-cudnn:{self.cudnn}-docker:{self.docker}"

    __repr__ = __str__


class BaseDependency(Protocol):
    def __init__(self, deps: t.Any) -> None:
        ...

    @property
    def kind(self) -> DependencyType:
        ...

    def conda_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        ...

    def venv_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        ...

    def asdict(self, ignore_keys: t.Optional[t.List[str]] = None) -> t.Dict:
        ...


@unique
class RuntimeRestoreStatus(Enum):
    restoring = "restoring"
    success = "success"
    failed = "failed"
    undefined = "undefined"

    @classmethod
    def mark(cls, status: RuntimeRestoreStatus, rootdir: Path) -> None:
        ensure_file(rootdir / ".runtime_restore_status", status.value, parents=True)

    @classmethod
    def get(cls, rootdir: Path) -> RuntimeRestoreStatus:
        fpath = rootdir / ".runtime_restore_status"
        if not fpath.exists():
            return cls.undefined
        else:
            try:
                return cls(fpath.read_text().strip())
            except ValueError:
                return cls.undefined


class NativeFileDependency(ASDictMixin, BaseDependency):
    def __init__(self, deps: t.List[t.Dict[str, str]]) -> None:
        if isinstance(deps, list):
            for d in deps:
                if not isinstance(d, dict):
                    raise FormatError(f"native file must dict type: {d}")
                if not d.get("src") or not d.get("dest"):
                    raise FormatError(
                        "dependencies.file MUST include src and dest fields."
                    )
        else:
            raise NoSupportError(
                f"native file dependency only support list[dict]: {deps}"
            )
        self.deps = deps

    @property
    def kind(self) -> DependencyType:
        return DependencyType.NATIVE_FILE

    def _do_install(self, src_dir: Path, mode: str) -> None:
        # TODO: support native file pre-hook, post-hook
        for d in self.deps:
            if in_container():
                _dest = Path(d["dest"])
            else:
                _mode_dir = src_dir / "export" / mode
                _dest = _mode_dir / d["dest"]
                if not is_within_dir(_mode_dir, _dest):
                    raise NoSupportError(
                        f"native files installation does not support the out of base({_mode_dir}) dir relative path({d['dest']}) in the host environment"
                    )

            # TODO: remove hard-code files dir
            _src = src_dir / RuntimeArtifactType.FILES / d["src"]
            console.print(f":baby_chick: copy native files: {_src} -> {_dest}")
            if _src.is_dir():
                copy_fs(str(_src), str(_dest))
            else:
                ensure_dir(_dest.parent)
                shutil.copyfile(str(_src), str(_dest))

    def conda_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        self._do_install(src_dir, PythonRunEnv.CONDA)

    def venv_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        self._do_install(src_dir, PythonRunEnv.VENV)


class WheelDependency(ASDictMixin, BaseDependency):
    def __init__(self, deps: t.List[str]) -> None:
        if isinstance(deps, list):
            for d in deps:
                if not isinstance(d, str) or not d.endswith(WHEEL_FILE_EXTENSION):
                    raise FormatError(f"wheel must str path: {d}")
        else:
            raise NoSupportError(f"wheel dependency only support list[str]: {deps}")
        self.deps = deps

    @property
    def kind(self) -> DependencyType:
        return DependencyType.WHEEL

    def _get_wheels(self, src_dir: Path) -> t.Generator[Path, None, None]:
        for d in self.deps:
            if not d:
                continue

            fpath = src_dir / d
            if not fpath.exists():
                raise NotFoundError(f"wheel install: {fpath}")

            yield fpath

    def conda_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        for path in self._get_wheels(src_dir):
            console.debug(f"conda run pip install: {path}")
            conda_install_req(req=path, prefix_path=env_dir, configs=configs)

    def venv_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        for path in self._get_wheels(src_dir):
            console.debug(f"venv pip install: {path}")
            venv_install_req(
                env_dir, path, pip_config=configs.get("pip")
            )  # type:ignore


class CondaPkgDependency(ASDictMixin, BaseDependency):
    def __init__(self, deps: t.List[str]) -> None:
        if isinstance(deps, list):
            for d in deps:
                if not isinstance(d, str):
                    raise FormatError(f"conda pkg must be str: {d}")
        else:
            raise NoSupportError(
                f"conda dependency only supports str or list[str] format: {deps}"
            )

        self.deps = deps

    @property
    def kind(self) -> DependencyType:
        return DependencyType.CONDA_PKG

    def conda_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        _conda_pkgs = " ".join([repr(_p) for _p in self.deps if _p])
        _conda_pkgs = _conda_pkgs.strip()
        if _conda_pkgs:
            console.debug(f"conda install: {_conda_pkgs}")
            conda_install_req(
                req=_conda_pkgs,
                prefix_path=env_dir,
                use_pip_install=False,
                configs=configs,
            )

    def venv_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        console.warning("no support install conda pkg in the venv environment")


class CondaEnvFileDependency(ASDictMixin, BaseDependency):
    def __init__(self, deps: str) -> None:
        if isinstance(deps, str):
            if not deps.endswith((".yaml", ".yml")):
                raise FormatError(
                    f"conda env file dependency must be .yaml or .yml file: {deps}"
                )
        else:
            raise NoSupportError(
                f"conda env file dependency only supports str format: {deps}"
            )

        self.deps = deps

    @property
    def kind(self) -> DependencyType:
        return DependencyType.CONDA_ENV_FILE

    def conda_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        env_fpath = src_dir / self.deps
        if not env_fpath.exists():
            raise NotFoundError(f"conda install env file: {env_fpath}")

        # TODO: configs for conda env update?
        conda_env_update(env_fpath=env_fpath, target_env=env_dir)

    def venv_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        console.warning(
            "no support install/update conda environment file in the venv environment"
        )


class PipPkgDependency(ASDictMixin, BaseDependency):
    def __init__(self, deps: t.List[str]) -> None:
        if isinstance(deps, list):
            for d in deps:
                if not isinstance(d, str):
                    raise FormatError(f"pip pkg must be str: {d}")
        else:
            raise NoSupportError(f"pip pkg dependency only supports str format: {deps}")

        self.deps = deps

    @property
    def kind(self) -> DependencyType:
        return DependencyType.PIP_PKG

    def _get_pkgs(self) -> t.Generator[str, None, None]:
        for d in self.deps:
            d = d.strip()
            if d:
                yield d

    def conda_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        # TODO: merge deps
        for pkg in self._get_pkgs():
            console.debug(f"conda run pip install: {pkg}")
            conda_install_req(req=pkg, prefix_path=env_dir, configs=configs)

    def venv_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        for pkg in self._get_pkgs():
            console.debug(f"venv pip install: {pkg}")
            venv_install_req(env_dir, pkg, pip_config=configs.get("pip"))  # type:ignore


class PipReqFileDependency(ASDictMixin, BaseDependency):
    def __init__(self, deps: str) -> None:
        if isinstance(deps, str):
            if not deps.endswith((".txt", ".in")):
                raise FormatError(
                    f"pip requirements file dependency must be .txt or .in file: {deps}"
                )
        else:
            raise NoSupportError(
                f"pip requirements file dependency only supports str format: {deps}"
            )

        self.deps = deps

    @property
    def kind(self) -> DependencyType:
        return DependencyType.PIP_REQ_FILE

    def _get_req_path(self, src_dir: Path) -> Path:
        # TODO: remove hard-code depend dir
        fpath = src_dir / self.deps
        if not fpath.exists():
            raise NotFoundError(f"pip req file: {fpath}")
        return fpath

    def conda_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        fpath = self._get_req_path(src_dir)
        conda_install_req(req=fpath, prefix_path=env_dir, configs=configs)

    def venv_install(self, src_dir: Path, env_dir: Path, configs: t.Dict) -> None:
        fpath = self._get_req_path(src_dir)
        venv_install_req(env_dir, fpath, pip_config=configs.get("pip"))  # type:ignore


dependency_map: t.Dict[DependencyType, t.Type[BaseDependency]] = {
    DependencyType.WHEEL: WheelDependency,
    DependencyType.CONDA_ENV_FILE: CondaEnvFileDependency,
    DependencyType.CONDA_PKG: CondaPkgDependency,
    DependencyType.PIP_PKG: PipPkgDependency,
    DependencyType.PIP_REQ_FILE: PipReqFileDependency,
    DependencyType.NATIVE_FILE: NativeFileDependency,
}


class Dependencies(ASDictMixin):
    def __init__(self, deps: t.Optional[t.List[str]] = None) -> None:
        deps = deps or []

        self.deps: t.List[BaseDependency] = []

        self._pip_pkgs: t.List[str] = []
        self._pip_files: t.List[str] = []
        self._conda_pkgs: t.List[str] = []
        self._conda_files: t.List[str] = []
        self._wheels: t.List[str] = []
        self._files: t.List[t.Dict[str, str]] = []
        self._unparsed: t.List[t.Any] = []

        _dmap: t.Dict[str, t.Tuple[t.Type[BaseDependency], t.List]] = {
            "pip": (PipPkgDependency, self._pip_pkgs),
            "conda": (CondaPkgDependency, self._conda_pkgs),
            "wheels": (WheelDependency, self._wheels),
            "files": (NativeFileDependency, self._files),
        }

        for d in deps:
            if isinstance(d, str):
                if d.endswith((".txt", ".in")):
                    self.deps.append(PipReqFileDependency(d))
                    self._pip_files.append(d)
                elif d.endswith((".yaml", ".yml")):
                    self.deps.append(CondaEnvFileDependency(d))
                    self._conda_files.append(d)
                else:
                    self._unparsed.append(d)
            elif isinstance(d, dict):
                for _k, _v in d.items():
                    if _k in _dmap:
                        _cls, _lst = _dmap[_k]
                        self.deps.append(_cls(_v))  # type: ignore
                        _lst.extend(_v)
                    else:
                        self._unparsed.append(d)
            else:
                self._unparsed.append(d)

        if self._unparsed:
            console.warning(f"unparsed dependencies:{self._unparsed}")

    def __str__(self) -> str:
        return f"Starwhale Runtime Dependencies: {len(self.deps)}, unparsed: {len(self._unparsed)}"

    def __repr__(self) -> str:
        return (
            f"Starwhale Runtime Dependencies: {len(self.deps)}, unparsed: {len(self._unparsed)}, pip pkg:{len(self._pip_pkgs)}"
            f"pip file:{len(self._pip_files)}, conda pkg:{len(self._conda_pkgs)} conda file:{len(self._conda_files)}"
            f"wheel: {len(self._wheels)} native file:{len(self._files)}"
        )

    def asdict(self, ignore_keys: t.Optional[t.List[str]] = None) -> t.Dict:
        return super().asdict(ignore_keys=ignore_keys or ["_unparsed"])

    def flatten_raw_deps(self) -> t.List:
        rt = []
        for d in self.deps:
            info = d.asdict()
            info["kind"] = d.kind.value
            rt.append(info)
        return rt


class Hooks(ASDictMixin):
    def __init__(self, pre: str = "", post: str = "", **kw: t.Any) -> None:
        self.pre = pre
        self.post = post


class DockerConfig(ASDictMixin):
    def __init__(
        self,
        image: str = "",
        **kw: t.Any,
    ) -> None:
        self.image = image


class PipConfig(ASDictMixin):
    def __init__(
        self,
        index_url: str = "",
        extra_index_url: _t_mixed_str_list = "",
        trusted_host: _t_mixed_str_list = "",
        **kw: t.Any,
    ) -> None:
        self.index_url = index_url
        self.extra_index_url = _list(extra_index_url)
        self.trusted_host = _list(trusted_host)


class CondaConfig(ASDictMixin):
    def __init__(self, channels: t.Optional[t.List[str]] = None, **kw: t.Any) -> None:
        self.channels = channels or [DEFAULT_CONDA_CHANNEL]


class Configs(ASDictMixin):
    def __init__(
        self,
        docker: t.Optional[t.Dict[str, str]] = None,
        conda: t.Optional[t.Dict[str, t.Any]] = None,
        pip: t.Optional[t.Dict[str, str]] = None,
        **kw: t.Any,
    ) -> None:
        self.docker = DockerConfig(**(docker or {}))
        self.conda = CondaConfig(**(conda or {}))
        self.pip = PipConfig(**(pip or {}))


class RuntimeConfig(ASDictMixin):
    def __init__(
        self,
        name: str,
        mode: str = PythonRunEnv.VENV,
        api_version: str = RUNTIME_API_VERSION,
        configs: t.Optional[t.Dict[str, t.Any]] = None,
        hooks: t.Optional[t.Dict[str, t.Any]] = None,
        dependencies: t.Optional[t.List[t.Any]] = None,
        environment: t.Optional[t.Dict[str, t.Any]] = None,
        python_version: str = "",
        pip_req: str = "",
        **kw: t.Any,
    ) -> None:
        self.name = name.strip().lower()
        self.mode = mode
        self.api_version = str(api_version)
        self.configs = Configs(**(configs or {}))
        self.hooks = Hooks(**(hooks or {}))

        environment = environment or {}
        if python_version and not environment.get("python"):
            environment["python"] = python_version
        self.environment = Environment(**environment)

        dependencies = dependencies or []
        if pip_req:
            dependencies.append(pip_req)
        self.dependencies = Dependencies(dependencies)

        self.kw = kw
        self._starwhale_version = get_downloadable_sw_version()
        self._do_validate()

    def _do_validate(self) -> None:
        ok, reason = validate_obj_name(self.name)
        if not ok:
            raise ConfigFormatError(f"name:{self.name}, reason:{reason}")

        if self.mode not in (PythonRunEnv.CONDA, PythonRunEnv.VENV):
            raise ConfigFormatError(f"{self.mode} no support")

        if self.api_version != RUNTIME_API_VERSION:
            raise NoSupportError(f"runtime api_version: {self.api_version}")

    @classmethod
    def create_by_yaml(cls, path: Path) -> RuntimeConfig:
        c = load_yaml(path)
        return cls(**c)

    def asdict(self, ignore_keys: t.Optional[t.List[str]] = None) -> t.Dict:
        return super().asdict(ignore_keys=ignore_keys or ["kw", "_starwhale_version"])


class Runtime(BaseBundle, metaclass=ABCMeta):
    @classmethod
    def restore(cls, workdir: Path, isolated_env_dir: t.Optional[Path] = None) -> None:
        StandaloneRuntime.restore(workdir, isolated_env_dir)

    @classmethod
    def get_runtime(cls, uri: Resource) -> Runtime:
        _cls = cls._get_cls(uri)
        return _cls(uri)

    @classmethod
    def _get_cls(  # type: ignore
        cls,
        uri: Resource,
    ) -> t.Union[t.Type[StandaloneRuntime], t.Type[CloudRuntime]]:
        if uri.instance.is_local:
            return StandaloneRuntime
        elif uri.instance.is_cloud:
            return CloudRuntime
        else:
            raise NoSupportError(f"runtime uri:{uri}")

    def __str__(self) -> str:
        return f"Starwhale Runtime: {self.uri}"

    @classmethod
    def copy(
        cls,
        src_uri: Resource,
        dest_uri: str,
        force: bool = False,
        dest_local_project_uri: str = "",
    ) -> None:
        bc = BundleCopy(
            src_uri,
            dest_uri,
            ResourceType.runtime,
            force,
            dest_local_project_uri=dest_local_project_uri,
        )
        bc.do()

    @classmethod
    def activate(cls, uri: Resource, force_restore: bool = False) -> None:
        StandaloneRuntime.activate(uri, force_restore)

    @classmethod
    def lock(
        cls,
        target_dir: str | Path,
        yaml_path: str | Path,
        env_name: str = "",
        env_prefix_path: str = "",
        no_cache: bool = False,
        stdout: bool = False,
        include_editable: bool = False,
        include_local_wheel: bool = False,
        emit_pip_options: bool = False,
        env_use_shell: bool = False,
    ) -> t.Tuple[str, t.Optional[Path]]:
        return StandaloneRuntime.lock(
            target_dir,
            yaml_path,
            env_name,
            env_prefix_path,
            no_cache,
            stdout,
            include_editable,
            include_local_wheel,
            emit_pip_options,
            env_use_shell,
        )

    def dockerize(
        self,
        tags: t.Optional[t.List[str]] = None,
        platforms: t.Optional[t.List[str]] = None,
        push: bool = True,
        dry_run: bool = False,
        use_starwhale_builder: bool = False,
        reset_qemu_static: bool = False,
    ) -> None:
        raise NotImplementedError

    def build_from_docker_image(self, image: str, runtime_name: str) -> None:
        raise NotImplementedError

    def build_from_python_env(
        self,
        runtime_name: str,
        mode: str = "",
        conda_name: str = "",
        conda_prefix: str = "",
        venv_prefix: str = "",
        cuda: str = "",
        cudnn: str = "",
        arch: str = "",
        download_all_deps: bool = False,
        include_editable: bool = False,
        include_local_wheel: bool = False,
    ) -> None:
        raise NotImplementedError

    def build_from_runtime_yaml(
        self,
        workdir: str | Path,
        yaml_path: str | Path,
        download_all_deps: bool = False,
        include_editable: bool = False,
        include_local_wheel: bool = False,
        no_cache: bool = False,
        disable_env_lock: bool = False,
        env_prefix_path: str = "",
        env_name: str = "",
        env_use_shell: bool = False,
    ) -> None:
        raise NotImplementedError


class StandaloneRuntime(Runtime, LocalStorageBundleMixin):
    def __init__(self, uri: Resource) -> None:
        super().__init__(uri)
        self.typ = InstanceType.STANDALONE
        self.store = RuntimeStorage(uri)
        self.tag = StandaloneTag(uri)
        self._manifest: t.Dict[str, t.Any] = {}
        self._version = uri.version
        self._detected_sw_version: str = ""

    def info(self) -> t.Dict[str, t.Any]:
        ret: t.Dict[str, t.Any] = {}
        if not self.store.bundle_path.exists():
            return ret

        ret["basic"] = {
            "name": self.uri.name,
            "uri": str(self.uri),
            "project": self.uri.project.name,
            "snapshot_workdir": str(self.store.snapshot_workdir),
            "bundle_path": str(self.store.bundle_path),
            "version": self.uri.version,
            "tags": StandaloneTag(self.uri).list(),
        }

        ret["basic"]["version"] = self.uri.version
        ret["basic"]["tags"] = StandaloneTag(self.uri).list()

        if self.store.snapshot_workdir.exists():
            ret["manifest"] = self.store.manifest
            ret["runtime_yaml"] = (
                self.store.snapshot_workdir / DefaultYAMLName.RUNTIME
            ).read_text()
            ret["lock"] = {}
            for fname in ret["manifest"]["environment"]["lock"]["files"]:
                ret["lock"][os.path.basename(fname)] = (
                    self.store.snapshot_workdir / "dependencies" / fname
                ).read_text()
        else:
            with TarFS(str(self.store.bundle_path)) as tar:
                with tar.open(DEFAULT_MANIFEST_NAME) as f:
                    ret["manifest"] = yaml.safe_load(f)

                with tar.open(DefaultYAMLName.RUNTIME) as f:
                    ret["runtime_yaml"] = f.read()

                ret["lock"] = {}
                for fname in ret["manifest"]["environment"]["lock"]["files"]:
                    fpath = os.path.join("dependencies", fname)
                    with tar.open(fpath) as f:
                        ret["lock"][os.path.basename(fpath)] = f.read()

        return ret

    def list_tags(self) -> t.List[str]:
        return self.tag.list()

    def add_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        self.tag.add(tags, ignore_errors)

    def remove_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        self.tag.remove(tags, ignore_errors)

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_remove(force)

    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: support short version to recover, today only support full-version
        dest_path = self.store.bundle_dir / f"{self.uri.version}{BundleType.RUNTIME}"
        _ok, _reason = move_dir(self.store.recover_loc, dest_path, force)
        _ok2, _reason2 = True, ""
        if self.store.recover_snapshot_workdir.exists():
            _ok2, _reason2 = move_dir(
                self.store.recover_snapshot_workdir, self.store.snapshot_workdir, force
            )
        return _ok and _ok2, _reason + _reason2

    def history(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.List[t.Dict[str, t.Any]]:
        # TODO: time order
        _r = []
        for _bf in self.store.iter_bundle_history():
            if not _bf.path.is_file():
                continue

            _r.append(
                dict(
                    version=_bf.version,
                    path=str(_bf.path.resolve()),
                    created_at=get_path_created_time(_bf.path),
                    size=_bf.path.stat().st_size,
                    tags=_bf.tags,
                )
            )
        return _r

    def build_from_docker_image(self, image: str, runtime_name: str) -> None:
        # TODO: validate image format
        workdir = Path(tempfile.mkdtemp(suffix="starwhale-runtime-build-"))
        try:
            config: t.Dict[str, t.Any] = dict(
                name=runtime_name,
                environment={
                    "docker": {"image": image},
                },
            )
            yaml_path = workdir / DefaultYAMLName.RUNTIME
            ensure_file(yaml_path, yaml.dump(config))
            self.build_from_runtime_yaml(
                workdir=workdir,
                yaml_path=yaml_path,
                disable_env_lock=True,
            )
        finally:
            empty_dir(workdir)

    def build_from_python_env(
        self,
        runtime_name: str,
        mode: str = PythonRunEnv.VENV,
        conda_name: str = "",
        conda_prefix: str = "",
        venv_prefix: str = "",
        cuda: str = "",
        cudnn: str = "",
        arch: str = "",
        download_all_deps: bool = False,
        include_editable: bool = False,
        include_local_wheel: bool = False,
    ) -> None:
        if conda_name or conda_prefix:
            prefix_path = (
                get_conda_prefix_path(conda_name) if conda_name else conda_prefix
            )
            if not check_valid_conda_prefix(prefix_path):
                raise RuntimeError(f"Invalid conda prefix: {prefix_path}")

            pybin = get_conda_pybin(prefix=prefix_path)
            env_use_shell = False
        elif venv_prefix:
            if not check_valid_venv_prefix(venv_prefix):
                raise RuntimeError(f"Invalid venv prefix: {venv_prefix}")
            pybin = os.path.join(venv_prefix, "bin", "python3")
            env_use_shell = False
        else:
            pybin = get_user_runtime_python_bin(mode)
            env_use_shell = True

        try:
            console.print(
                ":wheelchair: verify the installed packages have compatible dependencies"
            )
            pip_compatible_dependencies_check(pybin)
        except subprocess.CalledProcessError as e:
            console.print(
                f":skull: failed to verify compatible dependencies for the {pybin} environment: \n{e.output.strip()}",
                style="bold red",
            )
            sys.exit(1)

        python_version = get_python_version_by_bin(pybin)
        workdir = Path(tempfile.mkdtemp(suffix="starwhale-runtime-build-"))
        try:
            yaml_path = StandaloneRuntime.render_runtime_yaml(
                workdir,
                name=runtime_name,
                mode=mode,
                python_version=python_version,
                cuda_version=cuda,
                cudnn_version=cudnn,
                arch=arch,
            )

            self.build_from_runtime_yaml(
                workdir=workdir,
                yaml_path=yaml_path,
                download_all_deps=download_all_deps,
                include_editable=include_editable,
                include_local_wheel=include_local_wheel,
                disable_env_lock=False,
                env_name=conda_name,
                env_prefix_path=conda_prefix or venv_prefix,
                env_use_shell=env_use_shell,
            )
        finally:
            empty_dir(workdir)

    def build_from_runtime_yaml(
        self,
        workdir: str | Path,
        yaml_path: str | Path,
        download_all_deps: bool = False,
        include_editable: bool = False,
        include_local_wheel: bool = False,
        no_cache: bool = False,
        disable_env_lock: bool = False,
        env_prefix_path: str = "",
        env_name: str = "",
        env_use_shell: bool = False,
    ) -> None:
        workdir = Path(workdir)
        yaml_path = Path(yaml_path)
        swrt_config = self._load_runtime_config(yaml_path)

        lock_paths: t.List[Path] = []
        if not disable_env_lock:
            console.print(
                f":alien: try to lock environment dependencies to {workdir} ..."
            )
            content, lock_fpath = self.lock(
                target_dir=workdir,
                yaml_path=yaml_path,
                no_cache=no_cache,
                stdout=False,
                include_editable=include_editable,
                include_local_wheel=include_local_wheel,
                env_name=env_name,
                env_prefix_path=env_prefix_path,
                env_use_shell=env_use_shell,
            )
            # try getting starwhale version
            for line in content.splitlines():
                if line.startswith("starwhale=="):
                    self._detected_sw_version = line.split("==")[1].strip()
                    break

            if lock_fpath is not None:
                lock_paths.append(lock_fpath)

        operations = [
            (self._gen_version, 5, "gen version"),
            (self._prepare_snapshot, 5, "prepare snapshot"),
            (
                self._dump_context,
                5,
                "dump environment and configs",
                dict(config=swrt_config),
            ),
            (
                self._lock_environment,
                10,
                "lock environment",
                dict(
                    swrt_config=swrt_config,
                    env_prefix_path=env_prefix_path,
                    env_name=env_name,
                    env_use_shell=env_use_shell,
                    lock_paths=lock_paths,
                ),
            ),
            (
                self._dump_dependencies,
                50,
                "dump python dependencies",
                dict(
                    download_all_deps=download_all_deps,
                    deps=swrt_config.dependencies,
                    mode=swrt_config.mode,
                    include_editable=include_editable,
                    env_prefix_path=env_prefix_path,
                    env_name=env_name,
                ),
            ),
            (
                self._dump_base_image,
                5,
                "dump base image",
                dict(config=swrt_config),
            ),
            (
                self._copy_src,
                20,
                "dump src files:wheel, native files",
                dict(
                    config=swrt_config,
                    yaml_path=yaml_path,
                    lock_paths=lock_paths,
                ),
            ),
            (
                self._render_manifest,
                5,
                "render manifest",
            ),
            (self._make_tar, 20, "make runtime bundle", dict(ftype=BundleType.RUNTIME)),
            (self._make_auto_tags, 5, "make auto tags"),
        ]
        run_with_progress_bar("runtime bundle building...", operations)

    def _dump_context(self, config: RuntimeConfig) -> None:
        # TODO: refactor docker image in environment
        self._manifest["configs"] = config.configs.asdict()

    def _copy_src(
        self,
        config: RuntimeConfig,
        yaml_path: Path,
        lock_paths: t.Optional[t.List[Path]] = None,
    ) -> None:
        workdir = yaml_path.parent.resolve()
        yaml_name = yaml_path.name
        workdir_fs = open_fs(str(workdir))
        snapshot_fs = open_fs(str(self.store.snapshot_workdir.resolve()))
        copy_file(workdir_fs, yaml_name, snapshot_fs, DefaultYAMLName.RUNTIME)

        self._manifest["artifacts"] = {
            RuntimeArtifactType.RUNTIME: DefaultYAMLName.RUNTIME,
            RuntimeArtifactType.WHEELS: [],
            RuntimeArtifactType.DEPEND: [],
            RuntimeArtifactType.FILES: [],
        }

        console.info("[step:copy-wheels]start to copy wheels...")
        ensure_dir(self.store.snapshot_workdir / RuntimeArtifactType.WHEELS)
        for _fname in config.dependencies._wheels:
            _fpath = workdir / _fname
            if not _fpath.exists():
                console.warning(f"not found wheel: {_fpath}")
                continue

            _dest = f"{RuntimeArtifactType.WHEELS}/{_fname.lstrip('/')}"
            self._manifest["artifacts"][RuntimeArtifactType.WHEELS].append(_dest)
            copy_file(
                workdir_fs,
                _fname,
                snapshot_fs,
                _dest,
            )

        console.info("[step:copy-files]start to copy files...")
        ensure_dir(self.store.snapshot_workdir / RuntimeArtifactType.FILES)
        for _f in config.dependencies._files:
            _src = workdir / _f["src"]
            _dest = f"{RuntimeArtifactType.FILES}/{_f['src'].lstrip('/')}"
            if not _src.exists():
                console.warning(f"not found src-file: {_src}")
                continue

            self._manifest["artifacts"][RuntimeArtifactType.FILES].append(_f)
            # TODO: auto mkdir target parent dir?
            if _src.is_dir():
                # TODO: support .swignore file
                copy_fs(str(_src), str(self.store.snapshot_workdir / _dest))
            elif _src.is_file():
                ensure_dir((self.store.snapshot_workdir / _dest).parent)
                copy_file(workdir_fs, _f["src"], snapshot_fs, _dest)

        console.info("[step:copy-deps]start to copy pip/conda requirement files")
        ensure_dir(self.store.snapshot_workdir / RuntimeArtifactType.DEPEND)

        for _fname in config.dependencies._conda_files + config.dependencies._pip_files:
            _fpath = workdir / _fname
            if not _fpath.exists():
                console.warning(f"not found dependencies: {_fpath}")
                continue

            _dest = f"{RuntimeArtifactType.DEPEND}/{_fname.lstrip('/')}"
            ensure_dir((self.store.snapshot_workdir / _dest).parent)
            self._manifest["artifacts"][RuntimeArtifactType.DEPEND].append(_dest)
            copy_file(
                workdir_fs,
                _fname,
                snapshot_fs,
                _dest,
            )

        lock_paths = lock_paths or []
        for _fpath in lock_paths:
            _fname = _fpath.name
            _dest_name = f"{RuntimeArtifactType.DEPEND}/{SW_AUTO_DIRNAME}/lock/{_fname.lstrip('/')}"
            _dest_path = self.store.snapshot_workdir / _dest_name
            if _dest_path.exists():
                raise RuntimeError(
                    f"{_dest_name} has already been added into the runtime.yaml, failed to copy the auto-lock file"
                )

            ensure_dir(_dest_path.parent)
            self._manifest["artifacts"][RuntimeArtifactType.DEPEND].append(_dest_name)

            copy_file(
                open_fs(str(_fpath.parent)),
                _fname,
                snapshot_fs,
                _dest_name,
            )

    def _dump_base_image(self, config: RuntimeConfig) -> None:
        # prefer using image configured in runtime.yaml
        base_image = config.environment.docker.image

        if not base_image:
            _repo = os.environ.get(ENV_SW_IMAGE_REPO, DEFAULT_IMAGE_REPO)
            _tag = config._starwhale_version or LATEST_TAG
            base_image = SW_IMAGE_FMT.format(repo=_repo, tag=_tag)

            _cuda = config.environment.cuda
            _cudnn = config.environment.cudnn
            _suffix = []
            if _cuda:
                _suffix.append(f"-cuda{_cuda}")

                if _cudnn:
                    _suffix.append(f"-cudnn{_cudnn}")

            base_image += "".join(_suffix)

        console.print(
            f":rainbow: runtime docker image: [red]{base_image}[/]  :rainbow:"
        )
        self._manifest["base_image"] = base_image

    def _lock_environment(
        self,
        swrt_config: RuntimeConfig,
        env_prefix_path: str = "",
        env_name: str = "",
        env_use_shell: bool = False,
        lock_paths: t.Optional[t.List[Path]] = None,
    ) -> None:
        console.print(":bee: dump environment info...")
        sh_py_env = guess_current_py_env()
        sh_py_ver = get_user_python_version(sh_py_env)
        lock_paths = lock_paths or []

        self._manifest["environment"] = self._manifest.get("environment") or {}

        self._manifest["environment"].update(
            {
                "lock": {
                    "starwhale_version": self._detected_sw_version or STARWHALE_VERSION,
                    "system": platform.system(),
                    "shell": {
                        "python_env": sh_py_env,
                        "python_version": sh_py_ver,
                        "use_conda": is_conda(),
                        "use_venv": is_venv(),
                    },
                    "env_prefix_path": env_prefix_path,
                    "env_name": env_name,
                    "env_use_shell": env_use_shell,
                    "files": [f"{SW_AUTO_DIRNAME}/lock/{p.name}" for p in lock_paths],
                },
                "auto_lock_dependencies": bool(lock_paths),
                "python": swrt_config.environment.python,
                "arch": swrt_config.environment.arch,
                "mode": swrt_config.mode,
            }
        )

    def _dump_dependencies(
        self,
        mode: str = PythonRunEnv.AUTO,
        download_all_deps: bool = False,
        deps: t.Optional[Dependencies] = None,
        include_editable: bool = False,
        env_prefix_path: str = "",
        env_name: str = "",
    ) -> None:
        console.print(":bagel: dump dependencies info...")
        deps = deps or Dependencies()

        self._manifest["dependencies"] = {
            "raw_deps": deps.flatten_raw_deps(),
            "local_packaged_env": False,
            # compatibility with starwhale <= 0.3.0
            "pip_pkgs": deps._pip_pkgs,
            "conda_pkgs": deps._conda_pkgs,
            "pip_files": deps._pip_files,
            "conda_files": deps._conda_files,
        }

        console.info("[step:dep]finish dump dep")

        if download_all_deps:
            packaged = package_python_env(
                export_dir=self.store.export_dir,
                mode=mode,
                env_prefix_path=env_prefix_path,
                env_name=env_name,
                include_editable=include_editable,
            )
            self._manifest["dependencies"]["local_packaged_env"] = packaged

    def _prepare_snapshot(self) -> None:
        console.info("[step:prepare-snapshot]prepare runtime snapshot dirs...")

        # TODO: graceful clear?
        if self.store.snapshot_workdir.exists():
            raise ExistedError(str(self.store.snapshot_workdir))

        ensure_dir(self.store.snapshot_workdir)

        console.print(
            f":file_folder: workdir: [underline]{self.store.snapshot_workdir}[/]"
        )

    def _load_runtime_config(self, path: Path) -> RuntimeConfig:
        self._do_validate_yaml(path)
        return RuntimeConfig.create_by_yaml(path)

    def extract(self, force: bool = False, target: t.Union[str, Path] = "") -> Path:
        return self._do_extract(force, target)

    @classmethod
    def list(
        cls,
        project_uri: ProjectURI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.Union[t.Dict[str, t.Any], t.List[str]]] = None,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        filters = filters or {}
        rs = defaultdict(list)
        for _bf in RuntimeStorage.iter_all_bundles(
            project_uri, bundle_type=BundleType.RUNTIME, uri_type=ResourceType.runtime
        ):
            if not cls.do_bundle_filter(_bf, filters):
                continue

            if not _bf.path.is_file():
                continue

            # TODO: add more manifest info
            rs[_bf.name].append(
                {
                    "name": _bf.name,
                    "version": _bf.version,
                    "path": str(_bf.path.absolute()),
                    CREATED_AT_KEY: get_path_created_time(_bf.path),
                    "size": _bf.path.stat().st_size,
                    "is_removed": _bf.is_removed,
                    "tags": _bf.tags,
                }
            )
        return rs, {}

    @classmethod
    def quickstart_from_uri(
        cls,
        workdir: t.Union[Path, str],
        name: str,
        uri: Resource,
        force: bool = False,
        disable_restore: bool = False,
    ) -> None:
        workdir = Path(workdir).absolute()
        ensure_dir(workdir)

        if uri.instance.is_cloud:
            console.print(f":cloud: copy runtime from {uri} to local")
            _dest_project_uri = f"{STANDALONE_INSTANCE}/project/{DEFAULT_PROJECT}"
            cls.copy(uri, _dest_project_uri, force=force)
            uri = Resource(
                f"{_dest_project_uri}/runtime/{uri.name}/version/{uri.version}",
                typ=ResourceType.runtime,
            )

        sw_auto_d = workdir / SW_AUTO_DIRNAME
        extract_d = sw_auto_d / "fork-runtime-extract"
        console.print(f":package: extract swrt into {extract_d}...")
        _sr = StandaloneRuntime(uri)
        _sr.extract(force, extract_d)
        make_dir_gitignore(sw_auto_d)

        console.print(":printer: fork runtime files...")
        cls._do_fork_runtime_bundles(workdir, extract_d, name, force)

        if not disable_restore:
            console.print(f":safety_vest: start to restore to {extract_d}...")
            _manifest = load_yaml(extract_d / DEFAULT_MANIFEST_NAME)
            isolated_env_dir = sw_auto_d / _manifest["environment"]["mode"]
            if not isolated_env_dir.exists() or force:
                cls.restore(extract_d, isolated_env_dir)
            else:
                console.print(f":sake: {isolated_env_dir} existed, skip restore")

    @staticmethod
    def _do_fork_runtime_bundles(
        workdir: Path, extract_dir: Path, name: str, force: bool = False
    ) -> None:
        # TODO: support user define runtime.yaml name
        runtime_fpath = workdir / DefaultYAMLName.RUNTIME
        if (runtime_fpath).exists() and not force:
            raise ExistedError(runtime_fpath)

        def _copy_file(src: Path, dest: Path) -> None:
            if not src.exists():
                raise NotFoundError(src)

            if dest.exists() and not force:
                console.warning(f"{dest} existed, skip copy")

            ensure_dir(dest.parent)
            shutil.copy(str(src), str(dest))

        rt_config = RuntimeConfig.create_by_yaml(extract_dir / DefaultYAMLName.RUNTIME)
        wheel_dir = extract_dir / RuntimeArtifactType.WHEELS
        for _w in rt_config.dependencies._wheels:
            _copy_file(wheel_dir / _w, workdir / _w)

        dep_dir = extract_dir / RuntimeArtifactType.DEPEND
        for _d in (
            rt_config.dependencies._pip_files + rt_config.dependencies._conda_files
        ):
            _copy_file(dep_dir / _d, workdir / _d)

        files_dir = extract_dir / RuntimeArtifactType.FILES
        for _f in rt_config.dependencies._files:
            _copy_file(files_dir / _f["dest"], files_dir / _f["src"])

        runtime_yaml = load_yaml(extract_dir / DefaultYAMLName.RUNTIME)
        runtime_yaml["name"] = name
        ensure_file(
            runtime_fpath, yaml.safe_dump(runtime_yaml, default_flow_style=False)
        )

    @classmethod
    def quickstart_from_ishell(
        cls,
        workdir: t.Union[Path, str],
        name: str,
        mode: str,
        disable_create_env: bool = False,
        force: bool = False,
        interactive: bool = False,
    ) -> None:
        workdir = Path(workdir).absolute()
        ensure_dir(workdir)
        console.print(f":printer: render runtime.yaml @ {workdir}")
        python_version = get_python_version()

        sw_pkg = SW_PYPI_PKG_NAME
        if STARWHALE_VERSION and STARWHALE_VERSION != SW_DEV_DUMMY_VERSION:
            sw_pkg = f"{sw_pkg}=={STARWHALE_VERSION}"

        cls.render_runtime_yaml(
            workdir,
            name,
            mode,
            python_version,
            pkgs=[sw_pkg],
            force=True,
            auto_inject_sw=False,
        )

        if not disable_create_env:
            sw_auto_dir = workdir / SW_AUTO_DIRNAME
            make_dir_gitignore(sw_auto_dir)
            isolated_env_dir = sw_auto_dir / mode
            console.print(
                f":construction_worker: create {mode} isolated python environment..."
            )
            create_python_env(
                mode=mode,
                name=name,
                isolated_env_dir=isolated_env_dir,
                python_version=python_version,
                force=force,
            )
            console.print(f":dog: install {sw_pkg} {mode}@{isolated_env_dir}...")
            if mode == PythonRunEnv.VENV:
                venv_install_req(isolated_env_dir, req=sw_pkg, enable_pre=True)
            elif mode == PythonRunEnv.CONDA:
                conda_install_req(
                    prefix_path=isolated_env_dir, req=sw_pkg, enable_pre=True
                )

            activate_python_env(
                mode=mode,
                identity=str(isolated_env_dir.absolute()),
                interactive=interactive,
            )

    @classmethod
    def activate(cls, uri: Resource, force_restore: bool = False) -> None:
        if not uri.instance.is_local:
            raise NoSupportError(f"{uri} is not the standalone instance")

        _rt = StandaloneRuntime(uri)
        workdir = _rt.store.snapshot_workdir
        if not workdir.exists():
            console.print(f":package: extract swrt into {workdir}")
            _rt.extract(force=True, target=workdir)

        mode = load_yaml(_rt.store.manifest_path)["environment"]["mode"]
        prefix_path = _rt.store.export_dir / mode
        _rrs = RuntimeRestoreStatus
        is_invalid_status = _rrs.get(prefix_path) in (_rrs.failed, _rrs.restoring)

        if force_restore or not prefix_path.exists() or is_invalid_status:
            console.print(f":safety_vest: restore runtime into {workdir}")
            cls.restore(workdir)

        console.print(f":carrot: activate the current shell for the runtime uri: {uri}")
        activate_python_env(
            mode=mode, identity=str(prefix_path.resolve()), interactive=True
        )

    @classmethod
    def _ensure_isolated_python_env(
        cls,
        env_dir: Path,
        python_version: str,
        mode: str,
        no_cache: bool,
        recreate_env_if_broken: bool = True,
    ) -> None:
        if no_cache:
            empty_dir(env_dir)

        if env_dir.exists():
            is_valid_conda = mode == PythonRunEnv.CONDA and check_valid_conda_prefix(
                env_dir
            )
            is_valid_venv = mode == PythonRunEnv.VENV and check_valid_venv_prefix(
                env_dir
            )

            if is_valid_conda or is_valid_venv:
                return
            else:
                if recreate_env_if_broken:
                    empty_dir(env_dir)
                    ensure_dir(env_dir)
                else:
                    raise FormatError(f"{env_dir} is a valid {mode} dir")
        else:
            ensure_dir(env_dir)

        cls._setup_python_env(env_dir, mode=mode, python_version=python_version)

    @classmethod
    def lock(
        cls,
        target_dir: str | Path,
        yaml_path: str | Path,
        env_name: str = "",
        env_prefix_path: str = "",
        no_cache: bool = False,
        stdout: bool = False,
        include_editable: bool = False,
        include_local_wheel: bool = False,
        emit_pip_options: bool = False,
        env_use_shell: bool = False,
    ) -> t.Tuple[str, t.Optional[Path]]:
        """Install dependencies and save or print lock file

        Arguments:
            workdir: runtime work directory which contains the configured runtime.yaml
            yaml_name: runtime config file name
            env_name: conda environment name (used by conda env)
            env_prefix_path: python env prefix path (used by both venv and conda)
            no_cache: invalid all pkgs installed
            stdout: just print the lock info into stdout without saving lock file
            include_editable: include the editable pkg (only for venv)
            include_local_wheel: include local wheel pkg (only for venv)
            emit_pip_options: use user's pip configuration when freeze pkgs (only for venv)
            env_use_shell: automatically detect env in current shell session

        Returns:
            the lock file content(str) and the lock file path(Path)
        """
        yaml_path = Path(yaml_path)
        if not yaml_path.exists():
            raise NotFoundError(yaml_path)
        runtime_yaml = load_yaml(yaml_path)

        mode = runtime_yaml.get("mode", PythonRunEnv.VENV)
        expected_pyver = str(runtime_yaml.get("environment", {}).get("python", ""))
        _, temp_lock_path = tempfile.mkstemp(prefix="starwhale-lock-")
        console.print(f":butterfly: lock dependencies at mode {mode}")

        set_args = list(filter(bool, (env_name, env_prefix_path, env_use_shell)))
        if len(set_args) >= 2:
            raise ExclusiveArgsError(
                f"env_name({env_name}), env_prefix_path({env_prefix_path}) and env_use_shell({env_use_shell}) are the mutex arguments"
            )

        if env_name:
            if mode == PythonRunEnv.VENV:
                raise NoSupportError(
                    f"lock environment by the env name({env_name}) in venv mode"
                )
            prefix_path = get_conda_prefix_path(env_name)
        elif env_use_shell:
            prefix_path = get_base_prefix(mode)
        elif env_prefix_path:
            prefix_path = env_prefix_path
        else:
            _sw_auto_path = Path(target_dir) / SW_AUTO_DIRNAME / mode
            cls._ensure_isolated_python_env(
                _sw_auto_path, expected_pyver, mode, no_cache
            )
            prefix_path = str(_sw_auto_path)

        cls._install_dependencies_with_runtime_yaml(
            src_dir=yaml_path.parent,
            runtime_yaml=runtime_yaml,
            env_dir=prefix_path,
            skip_deps=[DependencyType.NATIVE_FILE],
        )

        if mode == PythonRunEnv.CONDA:
            if not check_valid_conda_prefix(prefix_path):
                raise FormatError(f"conda prefix: {prefix_path}")

            pybin = get_conda_pybin(prefix=prefix_path)
            console.print(
                f":cat_face: use conda env prefix({prefix_path}) to export environment..."
            )
            conda_export(
                temp_lock_path,
                prefix=prefix_path,
                include_editable=include_editable,
                include_local_wheel=include_local_wheel,
            )
        elif mode == PythonRunEnv.VENV:
            if not check_valid_venv_prefix(prefix_path):
                raise FormatError(f"venv prefix: {prefix_path}")

            pybin = os.path.join(str(prefix_path), "bin", "python3")
            console.print(f":cat_face: use {pybin} to freeze requirements...")
            pip_freeze_by_pybin(
                py_bin=pybin,
                lock_fpath=temp_lock_path,
                include_editable=include_editable,
                emit_options=emit_pip_options,
                include_local_wheel=include_local_wheel,
            )
        else:
            raise NoSupportError(f"lock {mode} environment")

        detected_pyver = get_python_version_by_bin(pybin)
        if expected_pyver and not detected_pyver.startswith(expected_pyver):
            raise EnvironmentError(
                f"{mode}: expected python({expected_pyver}) is not equal to detected python({detected_pyver})"
            )

        with open(temp_lock_path) as f:
            content = f.read()

        if stdout:
            console.rule("dependencies lock")
            console.print(content)
            os.unlink(temp_lock_path)
            return content, None
        else:
            dest_fname = (
                RuntimeLockFileType.CONDA
                if mode == PythonRunEnv.CONDA
                else RuntimeLockFileType.VENV
            )
            lock_dir = Path(target_dir) / SW_AUTO_DIRNAME / "lock"
            ensure_dir(lock_dir)
            dest_fpath = lock_dir / dest_fname
            shutil.move(temp_lock_path, dest_fpath)
            console.print(f":mouse: dump lock file: {dest_fpath}")
            return content, dest_fpath

    @staticmethod
    def render_runtime_yaml(
        workdir: t.Union[Path, str],
        name: str,
        mode: str,
        python_version: str = "",
        pkgs: t.Optional[t.List[str]] = None,
        force: bool = False,
        auto_inject_sw: bool = True,
        cuda_version: str = "",
        cudnn_version: str = "",
        arch: str = "",
    ) -> Path:
        workdir = Path(workdir)
        _rm = workdir / DefaultYAMLName.RUNTIME
        if _rm.exists() and not force:
            raise ExistedError(f"{_rm} was already existed")

        pkgs = pkgs or []
        if auto_inject_sw and not any(
            [p.strip().startswith(SW_PYPI_PKG_NAME) for p in pkgs]
        ):
            sw_pkg = SW_PYPI_PKG_NAME
            if STARWHALE_VERSION and STARWHALE_VERSION != SW_DEV_DUMMY_VERSION:
                sw_pkg = f"{sw_pkg}=={STARWHALE_VERSION}"
            pkgs.append(sw_pkg)
        pkgs = [p.strip() for p in pkgs if p.strip()]

        config: t.Dict[str, t.Any] = dict(
            name=name,
            mode=mode,
            environment={
                "python": python_version or get_python_version(),
                "arch": arch or SupportArch.NOARCH,
                "os": SupportOS.UBUNTU,
                "cuda": cuda_version,
                "cudnn": cudnn_version,
            },
            dependencies=[
                {"pip": pkgs},
            ],
            api_version=RUNTIME_API_VERSION,
        )
        ensure_file(_rm, yaml.safe_dump(config, default_flow_style=False), parents=True)
        return _rm

    def dockerize(
        self,
        tags: t.Optional[t.List[str]] = None,
        platforms: t.Optional[t.List[str]] = None,
        push: bool = True,
        dry_run: bool = False,
        use_starwhale_builder: bool = False,
        reset_qemu_static: bool = False,
    ) -> None:
        def _extract() -> None:
            if (
                not self.store.snapshot_workdir.exists()
                or not self.store.manifest_path.exists()
            ):
                console.print(":unicorn_face: extract runtime swrt file")
                self.extract(force=True)

        def _render_dockerfile(_manifest: t.Dict[str, t.Any]) -> None:
            console.print(f":wolf_face: render Dockerfile @{dockerfile_path}")
            _env = jinja2.Environment(
                loader=jinja2.FileSystemLoader(searchpath=_TEMPLATE_DIR)
            )
            _template = _env.get_template("Dockerfile.tmpl")
            _pip = _manifest["configs"].get("pip", {})
            _out = _template.render(
                base_image=_manifest["base_image"],
                runtime_name=self.uri.name,
                runtime_version=_manifest["version"],
                pypi_index_url=_pip.get("index_url", ""),
                pypi_extra_index_url=" ".join(_pip.get("extra_index_url", [])),
                pypi_trusted_host=" ".join(_pip.get("trusted_host", [])),
                python_version=_manifest["environment"]["python"],
                mode=_manifest["environment"]["mode"],
                local_packaged_env=_manifest["dependencies"].get(
                    "local_packaged_env", False
                ),
            )
            ensure_file(dockerfile_path, _out)

        def _render_dockerignore() -> None:
            _ignores = ["export/venv", "export/conda", ".git"]
            ensure_file(
                self.store.snapshot_workdir / ".dockerignore",
                content="\n".join(_ignores),
            )

        def _build(_manifest: t.Dict[str, t.Any]) -> None:
            _tags = list(tags or [])
            _platforms = platforms or []
            _dc_image = _manifest["configs"].get("docker", {}).get("image")
            if _dc_image:
                _tags.append(f"{_dc_image.split(':', 1)[0]}:{_manifest['version']}")

            if not dry_run and reset_qemu_static:
                docker.reset_qemu_static()

            docker.buildx(
                dockerfile_path,
                self.store.snapshot_workdir,
                push=push,
                dry_run=dry_run,
                platforms=_platforms,
                tags=_tags,
                use_starwhale_builder=use_starwhale_builder,
            )

        _extract()
        docker_dir = self.store.export_dir / "docker"
        ensure_dir(docker_dir)
        dockerfile_path = docker_dir / "Dockerfile"

        _manifest = load_yaml(self.store.manifest_path)
        _render_dockerfile(_manifest)
        _render_dockerignore()
        _build(_manifest)

    @classmethod
    def restore(
        cls,
        workdir: Path,
        isolated_env_dir: t.Optional[Path] = None,
        verbose: bool = True,
    ) -> None:
        if not (workdir.exists() and (workdir / DEFAULT_MANIFEST_NAME).exists()):
            raise NoSupportError("only support swrt extract workdir")

        _manifest = load_yaml(workdir / DEFAULT_MANIFEST_NAME)
        _env = _manifest["environment"]
        _starwhale_version = (
            _manifest.get("environment", {})
            .get("lock", {})
            .get("starwhale_version", "")
        )
        isolated_env_dir = isolated_env_dir or workdir / "export" / _env["mode"]

        operations: t.List[t.Any] = [
            (
                cls._validate_environment,
                5,
                "validate environment condition",
                dict(
                    expected_arch=_env.get("arch", []),
                ),
            ),
        ]

        if _manifest["dependencies"].get("local_packaged_env", False):
            operations.append(
                (
                    cls._extract_local_packaged_env,
                    20,
                    "extract local packaged env",
                    dict(
                        workdir=workdir,
                        mode=_env["mode"],
                        isolated_env_dir=isolated_env_dir,
                    ),
                )
            )
        else:
            operations.extend(
                [
                    (
                        cls._setup_python_env,
                        20,
                        "setup python env",
                        dict(
                            env_dir=isolated_env_dir
                            or (workdir / "export" / _env["mode"]),
                            mode=_env["mode"],
                            python_version=_env["python"],
                        ),
                    ),
                    (
                        cls._install_dependencies_within_restore,
                        50,
                        "install dependencies",
                        dict(
                            workdir=workdir,
                            mode=_env["mode"],
                            deps=_manifest["dependencies"],
                            configs=_manifest.get("configs", {}),
                            isolated_env_dir=isolated_env_dir,
                            lock_files=_env.get("lock", {}).get("files", []),
                        ),
                    ),
                ]
            )

        operations.extend(
            [
                (
                    install_starwhale,
                    10,
                    "install starwhale",
                    dict(
                        prefix_path=isolated_env_dir,
                        mode=_env["mode"],
                        version=_starwhale_version,
                        force=False,
                        configs=_manifest.get("configs", {}),
                    ),
                ),
                (
                    render_python_env_activate,
                    5,
                    "render python env activate scripts",
                    dict(
                        mode=_env["mode"],
                        prefix_path=isolated_env_dir,
                        workdir=workdir,
                        local_packaged_env=_manifest["dependencies"].get(
                            "local_packaged_env", False
                        ),
                        verbose=verbose,
                    ),
                ),
            ]
        )

        _status = RuntimeRestoreStatus
        _mark = partial(_status.mark, rootdir=isolated_env_dir)

        try:
            _mark(_status.restoring)
            run_with_progress_bar("runtime restore...", operations)
        except Exception:
            _mark(_status.failed)
            raise
        else:
            _mark(_status.success)

    @staticmethod
    def _install_dependencies_with_runtime_yaml(
        src_dir: Path,
        runtime_yaml: t.Any,
        env_dir: t.Union[Path, str],
        skip_deps: t.Optional[t.List[DependencyType]] = None,
    ) -> None:
        env_dir = Path(env_dir)

        mode = runtime_yaml.get("mode", PythonRunEnv.VENV)
        configs = runtime_yaml.get("configs", {})
        skip_deps = skip_deps or []
        console.print(
            f":baby_bottle: install runtime.yaml dependencies @ {env_dir} for lock..."
        )
        deps_config = Dependencies(runtime_yaml.get("dependencies", []))
        for dep in deps_config.deps:
            if dep.kind in skip_deps:
                console.debug(f"skip {dep} to install")
                continue

            _func = (
                dep.conda_install if PythonRunEnv.CONDA == mode else dep.venv_install
            )
            _func(src_dir, env_dir, configs)

    @staticmethod
    def _install_dependencies_within_restore(
        workdir: Path,
        mode: str,
        deps: t.Dict,
        configs: t.Dict,
        isolated_env_dir: t.Optional[Path] = None,
        lock_files: t.Optional[t.List[str]] = None,
    ) -> None:
        if "raw_deps" not in deps:
            raise NoSupportError(
                f"not found raw_deps field, please rebuild runtime with the newer swcli({STARWHALE_VERSION}) version"
            )

        lock_files = lock_files or []
        export_dir = workdir / "export"
        env_dir = isolated_env_dir or export_dir / mode

        src_map = {
            DependencyType.WHEEL: RuntimeArtifactType.WHEELS,
            DependencyType.CONDA_ENV_FILE: RuntimeArtifactType.DEPEND,
            DependencyType.PIP_REQ_FILE: RuntimeArtifactType.DEPEND,
        }

        if lock_files:
            # We assume the equation in the runtime auto-lock build mode:
            #   the lock files = pip_pkg + pip_req_file + conda_pkg + conda_env_file
            raw_deps = []
            for lf in lock_files:
                if lf.endswith(RuntimeLockFileType.CONDA):
                    raw_deps.append({"deps": lf, "kind": DependencyType.CONDA_ENV_FILE})
                elif lf.endswith(RuntimeLockFileType.VENV):
                    raw_deps.append({"deps": lf, "kind": DependencyType.PIP_REQ_FILE})
                else:
                    raise NoSupportError(
                        f"lock file({lf}) cannot be converted into raw_deps"
                    )

            # NATIVE_FILE and WHEEL must be installed after CONDA_ENV_FILE or PIP_REQ_FILE installation.
            for dep in deps["raw_deps"]:
                kind = DependencyType(dep["kind"])
                if kind in (DependencyType.NATIVE_FILE, DependencyType.WHEEL):
                    raw_deps.append(dep)
        else:
            raw_deps = deps["raw_deps"]

        for dep in raw_deps:
            kind = DependencyType(dep["kind"])
            if kind not in dependency_map:
                raise NoSupportError(f"install dependency:{kind}")

            console.print(
                f":dango: installing dependencies for [blink bold green]{kind.value}:{dep['deps']}[/] in the {mode} environment"
            )

            _obj = dependency_map[kind](dep["deps"])
            _func = (
                _obj.conda_install if PythonRunEnv.CONDA == mode else _obj.venv_install
            )
            if kind in src_map:
                src_dir = workdir / src_map[kind]
            else:
                src_dir = workdir

            _func(src_dir, env_dir, configs)

    @staticmethod
    def _setup_python_env(
        env_dir: Path,
        mode: str,
        python_version: str,
    ) -> None:
        console.print(f":abacus: setup python({python_version}) env with {mode}...")
        if mode == PythonRunEnv.CONDA:
            conda_setup(python_version, prefix=env_dir)
        elif mode == PythonRunEnv.VENV:
            venv_setup(env_dir, python_version=python_version)
        else:
            raise NoSupportError(f"ensure and build {mode} isolated python env")

    @staticmethod
    def _extract_local_packaged_env(
        workdir: Path,
        mode: str,
        isolated_env_dir: t.Optional[Path] = None,
    ) -> None:
        f = extract_conda_pkg if mode == PythonRunEnv.CONDA else extract_venv_pkg
        f(workdir, isolated_env_dir)

    @staticmethod
    def _validate_environment(expected_arch: t.List[str]) -> None:
        # TODO: add os, cuda version, python version validator
        def _validate_arch() -> None:
            machine_map = {
                "aarch64": SupportArch.ARM64,
                "arm64": SupportArch.ARM64,
                "x86_64": SupportArch.AMD64,
                "amd64": SupportArch.AMD64,
            }

            machine = platform.machine()
            machine_arch = [
                machine,
                machine_map.get(machine) or machine,
                SupportArch.NOARCH,
            ]
            _section = set([a.lower() for a in machine_arch]) & set(
                [a.lower() for a in expected_arch]
            )
            if not _section:
                raise UnExpectedConfigFieldError(
                    f"machine arch: {machine}, expected arch: {expected_arch}"
                )

        if expected_arch:
            _validate_arch()


class CloudRuntime(CloudBundleModelMixin, Runtime):
    def __init__(self, uri: Resource) -> None:
        super().__init__(uri)
        self.typ = InstanceType.CLOUD

    @classmethod
    @ignore_error(({}, {}))
    def list(
        cls,
        project_uri: ProjectURI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filter_dict: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        filter_dict = filter_dict or {}
        crm = CloudRequestMixed()
        return crm._fetch_bundle_all_list(
            project_uri, ResourceType.runtime, page, size, filter_dict
        )



from __future__ import annotations

import typing as t
from pathlib import Path

from rich import box
from rich.table import Table
from rich.pretty import Pretty

from starwhale.utils import console, pretty_bytes, pretty_merge_list
from starwhale.consts import DEFAULT_PAGE_IDX, DEFAULT_PAGE_SIZE, SHORT_VERSION_CNT
from starwhale.base.type import DatasetChangeMode
from starwhale.base.view import BaseTermView
from starwhale.base.uri.project import Project
from starwhale.base.uri.resource import Resource, ResourceType
from starwhale.core.dataset.type import DatasetConfig
from starwhale.core.runtime.process import Process as RuntimeProcess

from .model import Dataset


class DatasetTermView(BaseTermView):
    def __init__(self, dataset_uri: str | Resource) -> None:
        super().__init__()

        if isinstance(dataset_uri, Resource):
            self.uri = dataset_uri
        else:
            self.uri = Resource(dataset_uri, typ=ResourceType.dataset)
        self.dataset = Dataset.get_dataset(self.uri)

    @BaseTermView._simple_action_print
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.dataset.remove(force)

    @BaseTermView._simple_action_print
    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.dataset.recover(force)

    @BaseTermView._pager
    @BaseTermView._header
    def history(self, fullname: bool = False) -> t.List[t.Dict[str, t.Any]]:
        fullname = fullname or self.uri.instance.is_cloud
        return self._print_history(
            title="Dataset History List",
            history=self.dataset.history(),
            fullname=fullname,
        )

    def info(self) -> None:
        info = self.dataset.info()
        if info:
            console.print(Pretty(info, expand_all=True))
        else:
            console.print(f":bird: Dataset info not found: [bold red]{self.uri}[/]")

    def summary(self) -> None:
        summary = self.dataset.summary()
        if summary:
            console.print(Pretty(summary.asdict(), expand_all=True))
        else:
            console.print(":tea: not found dataset summary")

    def _do_diff(self, compare_uri: Resource) -> t.Dict[str, t.Any]:
        r = self.dataset.diff(compare_uri)
        r["diff_merged_output"] = {
            "added": pretty_merge_list(r["diff"]["added"]),
            "deleted": pretty_merge_list(r["diff"]["deleted"]),
            "updated": pretty_merge_list(
                [row["id"] for row, _ in r["diff"]["updated"]]
            ),
        }
        return r

    def diff(self, compare_uri: Resource, show_details: bool = False) -> None:
        _print_dict: t.Callable[[t.Dict], str] = lambda _s: "\n".join(
            [f"{k}:{v}" for k, v in _s.items()]
        )
        r = self._do_diff(compare_uri)
        table = Table(box=box.SIMPLE, expand=False, show_lines=True)
        table.add_column()
        table.add_column(
            f"base: {r['version']['base'][:SHORT_VERSION_CNT]}",
            style="magenta",
            justify="left",
        )
        table.add_column(
            f"compare: {r['version']['compare'][:SHORT_VERSION_CNT]}",
            style="cyan",
            justify="left",
        )

        table.add_row(
            "diff summary",
            _print_dict(r["diff_rows"]),
            style="red",
        )
        table.add_row(
            "dataset version",
            r["version"]["base"],
            r["version"]["compare"],
        )
        table.add_row(
            "dataset digest",
            _print_dict(r["summary"]["base"]),
            _print_dict(r["summary"]["compare"]),
        )

        def _print_diff(_diff: t.Union[str, list]) -> str:
            if isinstance(_diff, (list, tuple)):
                _diff = ",".join([str(_d) for _d in _diff])
            return _diff or "--"

        def _str_row(row: t.Dict) -> str:
            data_uri = (
                row["data_link"].uri[:SHORT_VERSION_CNT]
                if row["object_store_type"] == "local"
                else row["data_link"].uri
            )
            return f"{row['id']}:offset-{row['data_offset']}:size-{row['data_size']}:uri-{data_uri}"

        table.add_row("diff-deleted", _print_diff(r["diff_merged_output"]["deleted"]))
        table.add_row("diff-added", "", _print_diff(r["diff_merged_output"]["added"]))

        if show_details:
            _bout, _cout = [], []
            for _brow, _crow in r["diff"]["updated"]:
                _bout.append(_str_row(_brow))
                _cout.append(_str_row(_crow))

            table.add_row("diff-updated", "\n".join(_bout), "\n".join(_cout))
        else:
            table.add_row(
                "diff-updated", _print_diff(r["diff_merged_output"]["updated"])
            )

        console.print(table)

    @classmethod
    def list(
        cls,
        project_uri: t.Union[str, Project] = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.List[str]] = None,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        filters = filters or []
        if isinstance(project_uri, str):
            _uri = Project(project_uri)
        else:
            _uri = project_uri

        cls.must_have_project(_uri)
        fullname = fullname or _uri.instance.is_cloud
        _datasets, _pager = Dataset.list(_uri, page, size, filters)
        _data = BaseTermView.list_data(_datasets, show_removed, fullname)
        return _data, _pager

    @classmethod
    @BaseTermView._only_standalone
    def build(
        cls,
        workdir: str,
        config: DatasetConfig,
    ) -> None:
        if config.runtime_uri:
            RuntimeProcess(uri=config.runtime_uri).run()
        else:
            dataset_uri = cls.prepare_build_bundle(
                project=config.project_uri,
                bundle_name=config.name,
                typ=ResourceType.dataset,
                auto_gen_version=False,
            )
            ds = Dataset.get_dataset(dataset_uri)
            ds.build(workdir=Path(workdir), config=config)

    @classmethod
    def copy(
        cls,
        src_uri: str,
        dest_uri: str,
        mode: DatasetChangeMode = DatasetChangeMode.PATCH,
        dest_local_project_uri: str = "",
        force: bool = False,
    ) -> None:
        Dataset.copy(
            src_uri=Resource(src_uri, typ=ResourceType.dataset),
            dest_uri=dest_uri,
            mode=mode,
            dest_local_project_uri=dest_local_project_uri,
            force=force,
        )
        console.print(":clap: copy done")

    @BaseTermView._header
    def tag(
        self, tags: t.List[str], remove: bool = False, ignore_errors: bool = False
    ) -> None:
        if remove:
            console.print(f":golfer: remove tags {tags} @ {self.uri}...")
            self.dataset.remove_tags(tags, ignore_errors)
        else:
            console.print(f":surfer: add tags {tags} @ {self.uri}...")
            self.dataset.add_tags(tags, ignore_errors)

    @BaseTermView._header
    def head(self, rows: int, show_raw_data: bool = False) -> None:
        from starwhale.api._impl.data_store import _get_type

        for row in self.dataset.head(rows, show_raw_data):
            console.rule(f"row [{row['index']}]", align="left")
            output = (
                f":deciduous_tree: id: {row['index']} \n"
                ":cyclone: data:\n"
                f"\t :dim_button: type: {row['features']} \n"
            )
            for _k, _v in row["features"].items():
                ds_type: t.Any
                try:
                    ds_type = _get_type(_v)
                except RuntimeError:
                    ds_type = type(_v)
                output += (
                    f"\t :droplet: {_k}: value[{_v}], type[{ds_type} | {type(_v)}] \n"
                )

            console.print(output)


class DatasetTermViewRich(DatasetTermView):
    @classmethod
    @BaseTermView._pager
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.List[str]] = None,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        filters = filters or []
        _datasets, _pager = super().list(
            project_uri, fullname, show_removed, page, size, filters
        )
        custom_column: t.Dict[str, t.Callable[[t.Any], str]] = {
            "tags": lambda x: ",".join(x),
            "size": lambda x: pretty_bytes(x),
            "runtime": cls.place_holder_for_empty(),
        }

        cls.print_header()
        cls.print_table("Dataset List", _datasets, custom_column=custom_column)
        return _datasets, _pager


class DatasetTermViewJson(DatasetTermView):
    @classmethod
    def list(  # type: ignore
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.List[str]] = None,
    ) -> None:
        filters = filters or []
        _datasets, _pager = super().list(
            project_uri, fullname, show_removed, page, size, filters
        )
        cls.pretty_json(_datasets)

    def info(self) -> None:
        self.pretty_json(self.dataset.info())

    def head(self, rows: int, show_raw_data: bool = False) -> None:
        from starwhale.base.mixin import _do_asdict_convert

        info = self.dataset.head(rows, show_raw_data)
        self.pretty_json(_do_asdict_convert(info))

    def diff(self, compare_uri: Resource, show_details: bool = False) -> None:
        r = self._do_diff(compare_uri)
        if not show_details:
            r.pop("diff", None)

        self.pretty_json(r)

    def history(self, fullname: bool = False) -> None:
        fullname = fullname or self.uri.instance.is_cloud
        self.pretty_json(
            BaseTermView.get_history_data(self.dataset.history(), fullname)
        )


def get_term_view(ctx_obj: t.Dict) -> t.Type[DatasetTermView]:
    return (
        DatasetTermViewJson if ctx_obj.get("output") == "json" else DatasetTermViewRich
    )



from __future__ import annotations

import os
import typing as t
from pathlib import Path

import click
from click_option_group import (
    optgroup,
    MutuallyExclusiveOptionGroup,
    RequiredMutuallyExclusiveOptionGroup,
)

from starwhale.consts import DefaultYAMLName, DEFAULT_PAGE_IDX, DEFAULT_PAGE_SIZE
from starwhale.utils.cli import AliasedGroup
from starwhale.consts.env import SWEnv
from starwhale.utils.error import NoSupportError
from starwhale.core.model.view import get_term_view, ModelTermView
from starwhale.base.uri.project import Project
from starwhale.core.model.model import ModelConfig, ModelInfoFilter
from starwhale.core.model.store import ModelStorage
from starwhale.base.uri.resource import Resource, ResourceType
from starwhale.core.runtime.process import Process


@click.group(
    "model",
    cls=AliasedGroup,
    help="Model management, build/copy/run/list...",
)
@click.pass_context
def model_cmd(ctx: click.Context) -> None:
    ctx.obj = get_term_view(ctx.obj)


@model_cmd.command("build")
@click.argument("workdir", type=click.Path(exists=True, file_okay=False))
@click.option(
    "-p",
    "--project",
    default="",
    help="Project URI, default is the current selected project. The model package will store in the specified project.",
)
@click.option(
    "-f",
    "--model-yaml",
    default=None,
    help="mode yaml path, default use ${workdir}/model.yaml file",
)
@click.option(
    "-r",
    "--runtime",
    default="",
    help="runtime uri, build model in the runtime environment process",
)
@click.option(
    "modules",
    "-m",
    "--module",
    multiple=True,
    help="Python modules to be imported during the build process. The format is python module import path. The option supports set multiple times.",
)
@click.option("-n", "--name", default="", help="model name")
@click.option("-d", "--desc", default="", help="Dataset description")
@click.option(
    "--package-runtime/--no-package-runtime",
    is_flag=True,
    default=True,
    show_default=True,
    help="Package Starwhale Runtime into the model.",
)
@click.option(
    "--add-all",
    is_flag=True,
    default=False,
    help="Add all files in the working directory to the model package"
    "(excludes python cache files and virtual environment files when disabled)."
    "The '.swignore' file still takes effect.",
)
def _build(
    workdir: str,
    project: str,
    model_yaml: str,
    runtime: str,
    modules: t.List[str],
    package_runtime: bool,
    name: str,
    desc: str,
    add_all: bool,
) -> None:
    """Build starwhale model package.
    Only standalone instance supports model build.

    WORKDIR: model source code directory[Required].

    Example:

        \b
        # build by the model.yaml in current directory and model package will package all the files from the current directory.
        swcli model build .
        # search model run decorators from mnist.evaluate, mnist.train and mnist.predict modules, then package all the files from the current directory to model package.
        swcli model build . --module mnist.evaluate --module mnist.train --module mnist.predict
        # build model package in the Starwhale Runtime environment.
        swcli model build . --module mnist.evaluate --runtime pytorch/version/v1
        # forbid to package Starwhale Runtime into the model.
        swcli model build . --module mnist.evaluate --runtime pytorch/version/v1 --no-package-runtime
    """
    if model_yaml is None:
        yaml_path = Path(workdir) / DefaultYAMLName.MODEL
    else:
        yaml_path = Path(model_yaml)

    if yaml_path.exists():
        config = ModelConfig.create_by_yaml(yaml_path)
    else:
        config = ModelConfig()

    config.name = name or config.name or Path(workdir).name
    config.run.modules = modules or config.run.modules
    config.desc = desc
    config.do_validate()

    ModelTermView.build(
        workdir=workdir,
        project=project,
        model_config=config,
        runtime_uri=runtime,
        package_runtime=package_runtime,
        add_all=add_all,
    )


@model_cmd.command("extract")
@click.argument("model")
@click.argument("target_dir", default=".")
@click.option("-f", "--force", is_flag=True, help="Force to extract model package")
def _extract(model: str, target_dir: str, force: bool) -> None:
    """Extract model package to target directory.

    MODEL: model uri with version.

    TARGET_DIR: target directory to extract model package.

    Example:

        \b
        - extract mnist model package to current directory
            swcli model extract mnist/version/xxxx .

        - extract mnist model package to current directory and force to overwrite the files
            swcli model extract mnist/version/xxxx . -f
    """
    ModelTermView(model).extract(target=Path(target_dir), force=force)


@model_cmd.command("tag", help="Model Tag Management, add or remove")
@click.argument("model")
@click.argument("tags", nargs=-1)
@click.option("-r", "--remove", is_flag=True, help="Remove tags")
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help="Ignore tag name errors like name duplication, name absence",
)
def _tag(model: str, tags: t.List[str], remove: bool, quiet: bool) -> None:
    ModelTermView(model).tag(tags, remove, quiet)


@model_cmd.command("copy", aliases=["cp"])
@click.argument("src")
@click.argument("dest")
@click.option("-f", "--force", is_flag=True, help="Force to copy model")
@click.option("-dlp", "--dest-local-project", help="dest local project uri")
def _copy(src: str, dest: str, force: bool, dest_local_project: str) -> None:
    """
    Copy Model between Standalone Instance and Cloud Instance

    SRC: model uri with version

    DEST: project uri or model uri with name.

    Example:

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud model to local project(myproject) with a new model name 'mnist-local'
            swcli model cp cloud://pre-k8s/project/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq local/project/myproject/mnist-local

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud model to local default project(self) with the cloud instance model name 'mnist-cloud'
            swcli model cp cloud://pre-k8s/project/model/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq .

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud model to local project(myproject) with the cloud instance model name 'mnist-cloud'
            swcli model cp cloud://pre-k8s/project/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq . -dlp myproject

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud model to local default project(self) with a model name 'mnist-local'
            swcli model cp cloud://pre-k8s/project/model/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq mnist-local

        \b
        - copy cloud instance(pre-k8s) mnist project's mnist-cloud model to local project(myproject) with a model name 'mnist-local'
            swcli model cp cloud://pre-k8s/project/mnist/mnist-cloud/version/ge3tkylgha2tenrtmftdgyjzni3dayq mnist-local -dlp myproject

        \b
        - copy standalone instance(local) default project(self)'s mnist-local model to cloud instance(pre-k8s) mnist project with a new model name 'mnist-cloud'
            swcli model cp mnist-local/version/latest cloud://pre-k8s/project/mnist/mnist-cloud

        \b
        - copy standalone instance(local) default project(self)'s mnist-local model to cloud instance(pre-k8s) mnist project with standalone instance model name 'mnist-local'
            swcli model cp mnist-local/version/latest cloud://pre-k8s/project/mnist

        \b
        - copy standalone instance(local) default project(self)'s mnist-local model to cloud instance(pre-k8s) mnist project without 'cloud://' prefix
            swcli model cp mnist-local/version/latest pre-k8s/project/mnist

        \b
        - copy standalone instance(local) project(myproject)'s mnist-local model to cloud instance(pre-k8s) mnist project with standalone instance model name 'mnist-local'
            swcli model cp local/project/myproject/model/mnist-local/version/latest cloud://pre-k8s/project/mnist
    """
    ModelTermView.copy(src, dest, force, dest_local_project)


@model_cmd.command("info")
@click.argument("model")
@click.option(
    "-of",
    "--output-filter",
    type=click.Choice([f.value for f in ModelInfoFilter], case_sensitive=False),
    default=ModelInfoFilter.basic.value,
    show_default=True,
    help="Filter the output content. Only standalone instance supports this option.",
)
@click.pass_obj
def _info(view: t.Type[ModelTermView], model: str, output_filter: str) -> None:
    """Show model details.

    MODEL: argument use the `Model URI` format. Version is optional for the Model URI.
    If the version is not specified, the latest version will be used.

    Example:

        \b
        swcli model info mnist # show basic info from the latest version of model
        swcli model info mnist/version/v0 # show basic info from the v0 version of model
        swcli model info mnist/version/latest --output-filter=all # show all info
        swcli model info mnist -of basic # show basic info
        swcli model info mnist -of model_yaml  # show model.yaml
        swcli model info mnist -of handlers # show model runnable handlers info
        swcli model info mnist -of files # show model package files tree
        swcli -o json model info mnist -of all # show all info in json format
    """
    uri = Resource(model, typ=ResourceType.model)
    view(uri).info(ModelInfoFilter(output_filter))


@model_cmd.command("diff", help="model version diff")
@click.argument("base_uri", required=True)
@click.argument("compare_uri", required=True)
@click.option(
    "--show-details",
    is_flag=True,
    help="Show different detail by the model package files",
)
@click.pass_obj
def _diff(
    view: t.Type[ModelTermView], base_uri: str, compare_uri: str, show_details: bool
) -> None:
    view(base_uri).diff(Resource(compare_uri, typ=ResourceType.model), show_details)


@model_cmd.command("list", aliases=["ls"])
@click.option(
    "-p",
    "--project",
    default="",
    help="Project URI, default is the current selected project.",
)
@click.option("-f", "--fullname", is_flag=True, help="Show fullname of model version")
@click.option("-sr", "--show-removed", is_flag=True, help="Show removed model")
@click.option(
    "--page", type=int, default=DEFAULT_PAGE_IDX, help="Page number for model list"
)
@click.option(
    "--size", type=int, default=DEFAULT_PAGE_SIZE, help="Page size for model list"
)
@click.option(
    "filters",
    "-fl",
    "--filter",
    multiple=True,
    help="Filter output based on conditions provided.",
)
@click.pass_obj
def _list(
    view: t.Type[ModelTermView],
    project: str,
    fullname: bool,
    show_removed: bool,
    page: int,
    size: int,
    filters: t.List[str],
) -> None:
    """
    List Model of the specified project.

    The filtering flag (-fl or --filter) format is a key=value pair or a flag.
    If there is more than one filter, then pass multiple flags.\n
    (e.g. --filter name=mnist --filter latest)

    \b
    The currently supported filters are:
      name\tTEXT\tThe prefix of the model name
      owner\tTEXT\tThe name or id of the model owner
      latest\tFLAG\t[Cloud] Only show the latest version
            \t \t[Standalone] Only show the version with "latest" tag
    """
    view.list(project, fullname, show_removed, page, size, filters)


@model_cmd.command("history", help="Show model history")
@click.argument("model")
@click.option("--fullname", is_flag=True, help="Show version fullname")
@click.pass_obj
def _history(view: t.Type[ModelTermView], model: str, fullname: bool) -> None:
    view(model).history(fullname)


@model_cmd.command("remove", aliases=["rm"], help="Remove model")
@click.argument("model")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Force to remove model, the removed model cannot recover",
)
def _remove(model: str, force: bool) -> None:
    click.confirm("continue to remove?", abort=True)
    ModelTermView(model).remove(force)


@model_cmd.command("recover", help="Recover model")
@click.argument("model")
@click.option("-f", "--force", is_flag=True, help="Force to recover model")
def _recover(model: str, force: bool) -> None:
    ModelTermView(model).recover(force)


@model_cmd.command("run")
@optgroup.group(
    "\n ** Model Selectors",
    cls=RequiredMutuallyExclusiveOptionGroup,
    help="model uri or model source code dir",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-w",
    "--workdir",
    type=click.Path(exists=True, file_okay=False),
    help="Model source dir",
)
@optgroup.option("-u", "--uri", help="Model URI")  # type: ignore[no-untyped-call]
@optgroup.group(
    "\n ** Global Options",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-h",
    "--handler",
    default="0",
    show_default=True,
    help="runnable handler index or name, default is None, will use the first handler",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "modules",
    "-m",
    "--module",
    type=str,
    multiple=True,
    help="module name, the format is python module import path, handlers will be searched in the module. The option supports set multiple times.",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-f",
    "--model-yaml",
    help="Model yaml path, default use ${MODEL_DIR}/model.yaml file",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-p",
    "--run-project",
    envvar=SWEnv.project,
    default="",
    help=f"Project URI, env is {SWEnv.project}.The model run result will store in the specified project. Default is the current selected project.",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "datasets",
    "-d",
    "--dataset",
    required=False,
    envvar=SWEnv.dataset_uri,
    multiple=True,
    help=f"dataset uri, env is {SWEnv.dataset_uri}",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "--in-container",
    is_flag=True,
    help="[ONLY Standalone]Use docker container to run model handler, the docker image or runtime uri must be set",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-fs",
    "--forbid-snapshot",
    is_flag=True,
    help="[ONLY STANDALONE]Forbid to use model run snapshot dir, use model src dir directly",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "--cleanup-snapshot/--no-cleanup-snapshot",
    is_flag=True,
    default=True,
    show_default=True,
    help="[ONLY STANDALONE]Cleanup snapshot dir after model run",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "--resource-pool",
    default="default",
    type=str,
    help="resource pool for server side",
)
@optgroup.group(
    "\n ** Runtime Environment Selectors",
    cls=MutuallyExclusiveOptionGroup,
    help="The selector of runtime environment. If not set, model run in the current shell environment.",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-r",
    "--runtime",
    default="",
    help="runtime uri, model run in the runtime environment process",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-i",
    "--image",
    default="",
    help="[ONLY Standalone]docker image, works only when --use-docker is set",
)
@click.option(
    "--version",
    default=None,
    help="Run version",
    hidden=True,
)
@click.option("--step", default="", help="Evaluation run step", hidden=True)
@click.option(
    "--task-index",
    default=-1,
    type=int,
    help="Index of tasks in the current step",
    hidden=True,
)
@click.option(
    "--override-task-num",
    default=0,
    type=int,
    help="Total num of tasks in the current step",
    hidden=True,
)
@click.option(
    "-fpr",
    "--forbid-packaged-runtime",
    is_flag=True,
    help="[ONLY STANDALONE]Forbid to use packaged runtime in the model",
    hidden=True,
)
def _run(
    workdir: str,
    uri: str,
    handler: int | str,
    modules: t.List[str],
    model_yaml: str,
    run_project: str,
    datasets: t.List[str],
    in_container: bool,
    runtime: str,
    image: str,
    version: str | None,
    step: str,
    task_index: int | None,
    override_task_num: int,
    resource_pool: str,
    forbid_packaged_runtime: bool,
    forbid_snapshot: bool,
    cleanup_snapshot: bool,
) -> None:
    """Run Model.
    Model Package and the model source directory are supported.

    Examples:

        \b
        # --> run by model uri
        # run the first handler from model uri
        swcli model run -u mnist/version/latest
        # run index id(1) handler from model uri
        swcli model run --uri mnist/version/latest --handler 1
        # run index fullname(mnist.evaluator:MNISTInference.cmp) handler from model uri
        swcli model run --uri mnist/version/latest --handler mnist.evaluator:MNISTInference.cmp

        \b
        # --> run by the working directory, which does not build model package yet. Make local debug happy.
        # run the first handler from the working directory, use the model.yaml in the working directory
        swcli model run -w .
        # run index id(1) handler from the working directory, search mnist.evaluator module and model.yaml handlers(if existed) to get runnable handlers
        swcli model run --workdir . --module mnist.evaluator --handler 1
        # run index fullname(mnist.evaluator:MNISTInference.cmp) handler from the working directory, search mnist.evaluator module to get runnable handlers
        swcli model run --workdir . --module mnist.evaluator --handler mnist.evaluator:MNISTInference.cmp
    """
    # TODO: support run model in cluster mode
    run_project_uri = Project(run_project)
    in_server = run_project_uri.instance.is_cloud

    if in_container and in_server:
        raise RuntimeError("in-container and in-server are mutually exclusive")

    if in_server:
        if not runtime:
            raise ValueError("runtime is required in server mode")

        if not handler:
            raise ValueError("handler is required in server mode")

        if not uri:
            raise ValueError("uri is required in server mode")

        ModelTermView.run_in_server(
            project_uri=run_project_uri,
            model_uri=uri,
            dataset_uris=datasets,
            runtime_uri=runtime,
            resource_pool=resource_pool,
            run_handler=handler,
        )
        return

    model_src_dir, model_config, runtime_uri = _prepare_model_run_args(
        model=uri,
        runtime=runtime,
        workdir=workdir,
        modules=modules,
        model_yaml=model_yaml,
        forbid_packaged_runtime=forbid_packaged_runtime,
    )

    if in_container:
        ModelTermView.run_in_container(
            workdir=model_src_dir,
            runtime_uri=runtime_uri,
            docker_image=image,
        )
    else:
        ModelTermView.run_in_host(
            model_src_dir=model_src_dir,
            model_config=model_config,
            project=run_project,
            version=version,
            run_handler=handler,
            dataset_uris=datasets,
            runtime_uri=runtime_uri,
            forbid_snapshot=forbid_snapshot,
            cleanup_snapshot=cleanup_snapshot,
            scheduler_run_args={
                "step_name": step,
                "task_index": task_index,
                "task_num": override_task_num,
            },
            force_generate_jobs_yaml=uri is None,
        )


@model_cmd.command("serve")
@optgroup.group(
    "\n ** Model Selectors",
    cls=RequiredMutuallyExclusiveOptionGroup,
    help="model uri or model source code dir",
)
@optgroup.option(  # type: ignore[no-untyped-call]
    "-w",
    "--workdir",
    type=click.Path(exists=True, file_okay=False),
    help="Model source dir",
)
@optgroup.option("-u", "--uri", help="Model URI")  # type: ignore[no-untyped-call]
@click.option(  # type: ignore[no-untyped-call]
    "-f",
    "--model-yaml",
    help="Model yaml path, default use ${MODEL_DIR}/model.yaml file",
)
@click.option("-r", "--runtime", default="", help="runtime uri")
@click.option("--host", default="", help="The host to listen on")
@click.option("--port", default=8080, help="The port of the server")
@click.option(
    "-fpr",
    "--forbid-packaged-runtime",
    is_flag=True,
    help="[ONLY STANDALONE]Forbid to use packaged runtime in the model",
    hidden=True,
)
@click.option(
    "modules",
    "-m",
    "--module",
    multiple=True,
    help="Python modules to be imported during the build process. The format is python module import path. The option supports set multiple times.",
)
def _serve(
    workdir: str,
    uri: str,
    model_yaml: str,
    runtime: str,
    host: str,
    port: int,
    forbid_packaged_runtime: bool,
    modules: t.List[str],
) -> None:
    """Serve Model.

    Examples:

        \b
        swcli model serve -u mnist/version/latest
        swcli model serve --uri mnist/version/latest --runtime pytorch/version/latest

        \b
        swcli model serve --workdir . --runtime pytorch/version/latest
        swcli model serve --workdir . --runtime pytorch/version/latest --host 0.0.0.0 --port 8080
        swcli model serve --workdir . --runtime pytorch/version/latest --module mnist.evaluator
    """
    model_src_dir, config, runtime_uri = _prepare_model_run_args(
        model=uri,
        runtime=runtime,
        workdir=workdir,
        modules=modules,
        model_yaml=model_yaml,
        forbid_packaged_runtime=forbid_packaged_runtime,
    )

    ModelTermView.serve(
        model_src_dir=model_src_dir,
        model_config=config,
        host=host,
        port=port,
        runtime_uri=runtime_uri,
    )


def _prepare_model_run_args(
    model: str,
    runtime: str,
    workdir: str,
    modules: t.List[str],
    model_yaml: t.Optional[str],
    forbid_packaged_runtime: bool,
) -> t.Tuple[Path, ModelConfig, t.Optional[Resource]]:
    runtime_uri: Resource | None = None
    if runtime:
        try:
            runtime_uri = Resource(runtime, typ=ResourceType.runtime)
        except Exception:
            pass

    if model:
        model_uri = Resource(model, typ=ResourceType.model)
        model_store = ModelStorage(model_uri)
        model_src_dir = model_store.src_dir

        if modules:
            raise NoSupportError("module is not supported in model uri mode")

        if (
            os.environ.get(Process.EnvInActivatedProcess, "0") == "0"
            and model_store.digest.get("packaged_runtime")
            and not forbid_packaged_runtime
            and runtime_uri is None
        ):
            runtime_uri = model_uri
    else:
        model_src_dir = Path(workdir)

    model_src_dir = model_src_dir.absolute().resolve()
    if model_yaml is None:
        yaml_path = model_src_dir / DefaultYAMLName.MODEL
    else:
        yaml_path = Path(model_yaml)

    if yaml_path.exists():
        config = ModelConfig.create_by_yaml(yaml_path)
    else:
        config = ModelConfig()

    config.name = config.name or model_src_dir.name
    config.run.modules = modules or config.run.modules
    config.do_validate()

    return model_src_dir, config, runtime_uri



from __future__ import annotations

import os
import sys
import copy
import typing as t
from pathlib import Path
from functools import partial

from starwhale.utils import console
from starwhale.consts import PythonRunEnv
from starwhale.utils.fs import extract_tar
from starwhale.utils.venv import (
    get_conda_bin,
    guess_python_env_mode,
    check_valid_venv_prefix,
    check_valid_conda_prefix,
)
from starwhale.utils.error import NoSupportError, FieldTypeOrValueError
from starwhale.utils.process import check_call
from starwhale.core.model.model import StandaloneModel
from starwhale.base.uri.resource import Resource, ResourceType

from .model import StandaloneRuntime, RuntimeRestoreStatus


class Process:
    EnvInActivatedProcess = "SW_RUNTIME_ACTIVATED_PROCESS"
    ActivatedRuntimeURI = "SW_ACTIVATED_RUNTIME_URI_IN_SUBPROCESS"

    def __init__(
        self,
        uri: t.Union[Resource, str],
        force_restore: bool = False,
    ) -> None:
        self._uri = uri if isinstance(uri, Resource) else Resource(uri)
        self._prefix_path = self._restore_runtime(force_restore=force_restore)
        self._mode = guess_python_env_mode(self._prefix_path)

    def __str__(self) -> str:
        return f"process: {self._mode} in prefix path {self._prefix_path}"

    def __repr__(self) -> str:
        return f"process: {self._mode} in prefix path {self._prefix_path}, runtime uri: {self._uri}"

    def run(self) -> None:
        if os.environ.get(self.EnvInActivatedProcess, "0") == "1":
            raise RuntimeError("already in runtime activated process")

        argv = copy.deepcopy(sys.argv)
        if len(argv) < 2:
            raise NoSupportError("no cli command")

        argv[0] = f"{self._prefix_path}/bin/swcli"

        clear_positions = []
        # support formats: "--runtime=python3.7" or "--runtime python3.7" or "-r python3.7".
        # click lib does not support "-r=python3.7" format.
        for i in range(0, len(argv)):
            if argv[i] in ("-r", "--runtime"):
                clear_positions.append(i)
                clear_positions.append(i + 1)
                break
            elif argv[i].startswith("--runtime="):
                clear_positions.append(i)
                break

        if not clear_positions:
            raise RuntimeError(f"no runtime specified: {argv}")

        for p in clear_positions[::-1]:
            argv.pop(p)

        sub_cmd = " ".join(argv)

        # TODO: support windows platform
        if self._mode == PythonRunEnv.VENV:
            _bin_path = (self._prefix_path / "bin/activate").absolute()
            cmd = " && ".join([f"source {_bin_path}", sub_cmd])
        elif self._mode == PythonRunEnv.CONDA:
            cmd = f"{get_conda_bin()} run --live-stream --prefix {self._prefix_path.absolute()} {sub_cmd}"
        else:
            raise NoSupportError(f"get activate command for mode: {self._mode}")

        console.print(
            f":rooster: run process in the python isolated environment(prefix: {self._prefix_path})"
        )
        check_call(
            ["bash", "-c", cmd],
            env={
                self.EnvInActivatedProcess: "1",
                self.ActivatedRuntimeURI: str(self._uri),
            },
            log=partial(console.print, without_timestamp=True),
        )

    def _restore_runtime(
        self,
        force_restore: bool = False,
    ) -> Path:
        _uri = self._uri

        console.print(
            f":owl: start to run in the new process with runtime environment: {_uri}"
        )
        # TODO: support cloud runtime uri
        if _uri.instance.is_cloud:
            raise NoSupportError("run process with cloud instance uri")

        if _uri.typ == ResourceType.runtime:
            runtime = StandaloneRuntime(_uri)
            snapshot_workdir = runtime.store.snapshot_workdir
            bundle_path = runtime.store.bundle_path
        elif _uri.typ == ResourceType.model:
            model = StandaloneModel(_uri)
            snapshot_workdir = model.store.packaged_runtime_snapshot_workdir
            bundle_path = model.store.packaged_runtime_bundle_path
        else:
            raise FieldTypeOrValueError(
                f"{_uri} is not a valid uri, only support model(packaged runtime) or runtime uri"
            )

        venv_prefix = snapshot_workdir / "export" / PythonRunEnv.VENV
        conda_prefix = snapshot_workdir / "export" / PythonRunEnv.CONDA

        is_valid_prefix = check_valid_venv_prefix(
            venv_prefix
        ) or check_valid_conda_prefix(conda_prefix)

        _rrs = RuntimeRestoreStatus
        is_invalid_status = set([_rrs.get(venv_prefix), _rrs.get(conda_prefix)]) & set(
            [_rrs.failed, _rrs.restoring]
        )

        if force_restore or not is_valid_prefix or is_invalid_status:
            console.print(f":snail: start to restore runtime: {_uri}")
            extract_tar(tar_path=bundle_path, dest_dir=snapshot_workdir, force=True)
            StandaloneRuntime.restore(snapshot_workdir, verbose=False)

        if venv_prefix.exists():
            prefix = venv_prefix
        elif conda_prefix.exists():
            prefix = conda_prefix
        else:
            raise RuntimeError(
                f"venv_prefix({venv_prefix}) and conda_prefix({conda_prefix}) are both not existed"
            )

        return prefix.resolve()



import typing as t
from pathlib import Path

from starwhale.utils import load_yaml
from starwhale.consts import (
    SW_AUTO_DIRNAME,
    DIGEST_FILE_NAME,
    VERSION_PREFIX_CNT,
    RESOURCE_FILES_NAME,
    DEFAULT_MANIFEST_NAME,
    DEFAULT_SW_TASK_RUN_IMAGE,
)
from starwhale.base.type import BundleType
from starwhale.base.store import BaseStorage
from starwhale.base.uri.resource import ResourceType


class ModelStorage(BaseStorage):
    def _guess(self) -> t.Tuple[Path, str]:
        return self._guess_for_bundle()

    @property
    def bundle_type(self) -> str:
        return BundleType.MODEL

    @property
    def uri_type(self) -> str:
        return ResourceType.model.value

    @property
    def src_dir_name(self) -> str:
        return "src"

    @property
    def src_dir(self) -> Path:
        return self.snapshot_workdir / self.src_dir_name

    @property
    def hidden_sw_dir(self) -> Path:
        return self.src_dir / SW_AUTO_DIRNAME

    @property
    def recover_loc(self) -> Path:
        return self._get_recover_loc_for_bundle()

    @property
    def resource_files_path(self) -> Path:
        return self.hidden_sw_dir / RESOURCE_FILES_NAME

    @property
    def resource_files(self) -> t.List[t.Dict[str, t.Any]]:
        if not self.resource_files_path.exists():
            return self.manifest.get("resources", [])
        else:
            return load_yaml(self.resource_files_path)  # type: ignore

    @property
    def digest_path(self) -> Path:
        return self.hidden_sw_dir / DIGEST_FILE_NAME

    @property
    def digest(self) -> t.Dict[str, t.Any]:
        if not self.digest_path.exists():
            _manifest = self.manifest
            _manifest.pop("resources", None)
            _manifest.pop("build", None)
            return _manifest
        else:
            return load_yaml(self.digest_path)  # type: ignore

    @property
    def manifest_path(self) -> Path:
        return self.snapshot_workdir / DEFAULT_MANIFEST_NAME

    @property
    def snapshot_workdir(self) -> Path:
        if self.building:
            return self.tmp_dir
        version = self.uri.version
        return (
            self.project_dir
            / self.uri.typ.value
            / self.uri.name
            / version[:VERSION_PREFIX_CNT]
            / f"{version}{BundleType.MODEL}"
        )

    @property
    def packaged_runtime_snapshot_workdir(self) -> Path:
        return self.snapshot_workdir / "runtime_snapshot_workdir"

    @property
    def packaged_runtime_export_dir(self) -> Path:
        return self.packaged_runtime_snapshot_workdir / "export"

    @property
    def packaged_runtime_bundle_path(self) -> Path:
        return self.hidden_sw_dir / "runtime" / f"packaged{BundleType.RUNTIME}"

    def get_packaged_runtime_base_image(self) -> str:
        manifest = self.get_manifest_by_path(
            fpath=self.packaged_runtime_bundle_path,
            bundle_type=BundleType.RUNTIME,
            uri_type=ResourceType.runtime.value,
        )
        return manifest.get("base_image", DEFAULT_SW_TASK_RUN_IMAGE)



from __future__ import annotations

import os
import typing as t
from pathlib import Path

from rich import box
from rich.text import Text
from rich.tree import Tree
from rich.panel import Panel
from rich.table import Table
from rich.pretty import Pretty
from rich.syntax import Syntax
from rich.console import Group

from starwhale.utils import docker, console, process, pretty_bytes, in_production
from starwhale.consts import (
    FileFlag,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    SHORT_VERSION_CNT,
)
from starwhale.utils.fs import cmp_file_content
from starwhale.base.view import BaseTermView
from starwhale.consts.env import SWEnv
from starwhale.utils.error import NoSupportError, FieldTypeOrValueError
from starwhale.base.uri.project import Project
from starwhale.core.model.store import ModelStorage
from starwhale.base.uri.resource import Resource, ResourceType
from starwhale.core.runtime.model import StandaloneRuntime
from starwhale.core.runtime.process import Process as RuntimeProcess

from .model import Model, CloudModel, ModelConfig, ModelInfoFilter, StandaloneModel


class ModelTermView(BaseTermView):
    def __init__(self, model_uri: str | Resource) -> None:
        super().__init__()

        if isinstance(model_uri, Resource):
            self.uri = model_uri
        else:
            self.uri = Resource(model_uri, ResourceType.model, refine=True)
        self.model = Model.get_model(self.uri)

    @BaseTermView._simple_action_print
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.model.remove(force)

    @BaseTermView._simple_action_print
    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.model.recover(force)

    def info(self, output_filter: ModelInfoFilter = ModelInfoFilter.basic) -> None:
        info = self.model.info()
        if not info:
            console.print(f":bird: model info not found: [bold red]{self.uri}[/]")
            return

        def _render_handlers() -> Table:
            table = Table(
                title="Model Package Runnable Handlers",
                box=box.MARKDOWN,
                show_lines=True,
                expand=True,
            )
            table.add_column("Handler Index", style="cyan")
            table.add_column("Handler Name", style="cyan")
            table.add_column("Steps", style="green")

            handlers = sorted(info.get("handlers", {}).items())
            for index, (name, steps) in enumerate(handlers):
                steps_content = []
                for step in steps:
                    steps_content.append(f":palm_tree: {step['name']}")
                    steps_content.append(f"\t - replicas: {step['replicas']}")
                    steps_content.append(f"\t - needs: {' '.join(step['needs'])}")
                table.add_row(str(index), name, "\n".join(steps_content))
            return table

        def _render_files_tree() -> Tree:
            files = info.get("files", [])
            unfold_files: t.Dict[str, t.Any] = {}
            for f in files:
                path = Path(f["path"])
                parent = unfold_files
                for p in path.parent.parts:
                    parent.setdefault(p, {})
                    parent = parent[p]

                parent[path.name] = f"{pretty_bytes(f['size'])}"

            root = Tree("Model Resource Files Tree", style="bold bright_blue")

            def _walk(tree: Tree, files: t.Dict) -> None:
                for k, v in files.items():
                    if isinstance(v, dict):
                        subtree = tree.add(f":file_folder: {k}")
                        _walk(subtree, v)
                    else:
                        tree.add(f":page_facing_up: {k} ({v})")

            _walk(root, unfold_files)
            return root

        basic_content = Pretty(info.get("basic", {}), expand_all=True)
        model_yaml_content = Syntax(
            info.get("model_yaml", ""), "yaml", theme="ansi_dark"
        )
        handlers_content = _render_handlers()
        files_content = _render_files_tree()

        # TODO: support files tree
        if output_filter == ModelInfoFilter.basic:
            console.print(basic_content)
        elif output_filter == ModelInfoFilter.model_yaml:
            console.print(model_yaml_content)
        elif output_filter == ModelInfoFilter.handlers:
            console.print(handlers_content)
        elif output_filter == ModelInfoFilter.files:
            console.print(files_content)
        else:
            console.rule("[green bold] Model Basic Info")
            console.print(basic_content)
            console.rule("[green bold] model.yaml")
            console.print(model_yaml_content)
            console.rule("[green bold] Model Handlers")
            console.print(handlers_content)

    @BaseTermView._only_standalone
    def extract(self, target: Path, force: bool = False) -> None:
        self.model.extract(target=target, force=force)

    @BaseTermView._only_standalone
    def diff(self, compare_uri: Resource, show_details: bool) -> None:
        r = self.model.diff(compare_uri)
        text_details: t.List[Panel] = []
        table = Table(box=box.SIMPLE, expand=False, show_lines=True)
        table.add_column("file")
        table.add_column("path")
        table.add_column(
            f"base: {self.model.version[:SHORT_VERSION_CNT]}",
            style="magenta",
            justify="left",
        )
        table.add_column(
            f"compare: {compare_uri.version[:SHORT_VERSION_CNT]}",
            style="cyan",
            justify="left",
        )
        b_files = r["base_version"]
        c_files = r["compare_version"]
        for _p in r["all_paths"]:
            _flag = c_files[_p].flag
            if _flag == FileFlag.ADDED:
                _flag_txt = f"[blue]{_flag}"
            elif _flag == FileFlag.DELETED:
                _flag_txt = f"[red]{_flag}"
            elif _flag == FileFlag.UPDATED:
                _flag_txt = f"[yellow]{_flag}"
                if show_details:
                    text_details.append(
                        Panel(
                            Text(
                                text="".join(
                                    cmp_file_content(
                                        base_path=b_files[_p].path,
                                        cmp_path=c_files[_p].path,
                                    )
                                )
                            ),
                            title=_p,
                        )
                    )
            else:
                _flag_txt = f"[green]{_flag}"
            table.add_row(
                os.path.basename(_p),
                _p,
                "existed" if _p in b_files else "-",
                f"[green]{_flag_txt}",
            )

        console.print(Panel(table, title="diff overview"))
        if show_details:
            console.print(Panel(Group(*text_details), title="diff details"))

    @BaseTermView._pager
    @BaseTermView._header
    def history(self, fullname: bool = False) -> t.List[t.Dict[str, t.Any]]:
        fullname = fullname or self.uri.instance.is_cloud
        return self._print_history(
            title="Model History List", history=self.model.history(), fullname=fullname
        )

    @classmethod
    def run_in_server(
        cls,
        project_uri: Project,
        model_uri: str,
        dataset_uris: t.List[str],
        runtime_uri: str,
        resource_pool: str,
        run_handler: str | int,
    ) -> t.Tuple[bool, str]:
        ok, version_or_reason = CloudModel.run(
            project_uri=project_uri,
            model_uri=model_uri,
            dataset_uris=dataset_uris,
            runtime_uri=runtime_uri,
            resource_pool=resource_pool,
            run_handler=run_handler,
        )
        if ok:
            console.print(":clap: success to create job")
            console.print(
                f":bird: visit web: {project_uri.instance.url}/projects/{project_uri.name}/evaluations/{version_or_reason}"
            )
            console.print(
                f":monkey: run command: [bold green]swcli job info {project_uri.full_uri}/job/{version_or_reason} [/]"
            )
        else:
            console.print(f":bird: run failed: [bold red]{version_or_reason}[/]")

        return ok, version_or_reason

    @classmethod
    @BaseTermView._only_standalone
    def run_in_host(
        cls,
        model_src_dir: Path | str,
        model_config: ModelConfig,
        project: str = "",
        version: str = "",
        run_handler: str = "",
        dataset_uris: t.Optional[t.List[str]] = None,
        runtime_uri: t.Optional[Resource] = None,
        scheduler_run_args: t.Optional[t.Dict] = None,
        forbid_snapshot: bool = False,
        cleanup_snapshot: bool = True,
        force_generate_jobs_yaml: bool = False,
    ) -> None:
        if runtime_uri:
            RuntimeProcess(uri=runtime_uri).run()
        else:
            StandaloneModel.run(
                model_src_dir=Path(model_src_dir),
                model_config=model_config,
                project=Project(project).name,
                version=version,
                run_handler=run_handler,
                dataset_uris=dataset_uris,
                scheduler_run_args=scheduler_run_args,
                forbid_snapshot=forbid_snapshot,
                cleanup_snapshot=cleanup_snapshot,
                force_generate_jobs_yaml=force_generate_jobs_yaml,
            )

    @classmethod
    @BaseTermView._only_standalone
    def run_in_container(
        cls,
        model_src_dir: Path,
        runtime_uri: t.Optional[Resource] = None,
        docker_image: str = "",
    ) -> None:
        # TODO: support to get job version for in container
        if not runtime_uri and not docker_image:
            raise FieldTypeOrValueError("runtime_uri and docker_image both are none")

        if runtime_uri:
            if runtime_uri.typ == ResourceType.runtime:
                runtime = StandaloneRuntime(runtime_uri)
                docker_image = runtime.store.get_docker_base_image()
            elif runtime_uri.typ == ResourceType.model:
                model = StandaloneModel(runtime_uri)
                docker_image = model.store.get_packaged_runtime_base_image()
            else:
                raise NoSupportError(f"not support {runtime_uri} to fetch docker image")

        mounts = [str(model_src_dir.resolve().absolute())]
        envs = {SWEnv.runtime_version: str(runtime_uri)} if runtime_uri else {}
        cmd = docker.gen_swcli_docker_cmd(
            docker_image,
            envs=envs,
            mounts=mounts,
        )
        console.rule(":elephant: docker cmd", align="left")
        console.print(f"{cmd}\n")
        process.check_call(cmd, shell=True)

    @classmethod
    def _get_workdir(cls, target: str) -> Path:
        if in_production() or (os.path.exists(target) and os.path.isdir(target)):
            workdir = Path(target)
        else:
            _uri = Resource(target, ResourceType.model)
            _store = ModelStorage(_uri)
            workdir = _store.src_dir
        return workdir

    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.Union[t.Dict[str, t.Any], t.List[str]]] = None,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        filters = filters or {}
        _uri = Project(project_uri)
        cls.must_have_project(_uri)
        fullname = fullname or _uri.instance.is_cloud
        _models, _pager = Model.list(_uri, page, size, filters)
        _data = BaseTermView.list_data(_models, show_removed, fullname)
        return _data, _pager

    @classmethod
    @BaseTermView._only_standalone
    def build(
        cls,
        workdir: t.Union[str, Path],
        project: str,
        model_config: ModelConfig,
        add_all: bool,
        runtime_uri: str = "",
        package_runtime: bool = False,
    ) -> None:
        if runtime_uri:
            RuntimeProcess(uri=Resource(runtime_uri, typ=ResourceType.runtime)).run()
        else:
            model_uri = cls.prepare_build_bundle(
                project=project,
                bundle_name=model_config.name,
                typ=ResourceType.model,
            )
            m = Model.get_model(model_uri)

            if package_runtime:
                packaging_runtime_uri = os.environ.get(
                    RuntimeProcess.ActivatedRuntimeURI
                )
            else:
                packaging_runtime_uri = None

            m.build(
                Path(workdir),
                model_config=model_config,
                packaging_runtime_uri=packaging_runtime_uri,
                add_all=add_all,
            )

    @classmethod
    def copy(
        cls,
        src_uri: str,
        dest_uri: str,
        force: bool = False,
        dest_local_project_uri: str = "",
    ) -> None:
        src = Resource(src_uri, typ=ResourceType.model)
        Model.copy(src, dest_uri, force, dest_local_project_uri)
        console.print(":clap: copy done.")

    @BaseTermView._header
    def tag(
        self, tags: t.List[str], remove: bool = False, ignore_errors: bool = False
    ) -> None:
        if remove:
            console.print(f":golfer: remove tags [red]{tags}[/] @ {self.uri}...")
            self.model.remove_tags(tags, ignore_errors)
        else:
            console.print(f":surfer: add tags [red]{tags}[/] @ {self.uri}...")
            self.model.add_tags(tags, ignore_errors)

    @classmethod
    @BaseTermView._only_standalone
    def serve(
        cls,
        model_src_dir: Path,
        model_config: ModelConfig,
        runtime_uri: Resource,
        host: str,
        port: int,
    ) -> None:
        if runtime_uri:
            RuntimeProcess(uri=runtime_uri).run()
        else:
            StandaloneModel.serve(
                model_config=model_config,
                model_src_dir=model_src_dir,
                host=host,
                port=port,
            )


class ModelTermViewRich(ModelTermView):
    @classmethod
    @BaseTermView._pager
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.List[str]] = None,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        filters = filters or []
        _models, _pager = super().list(
            project_uri, fullname, show_removed, page, size, filters
        )
        custom_column: t.Dict[str, t.Callable[[t.Any], str]] = {
            "tags": lambda x: ",".join(x),
            "size": lambda x: pretty_bytes(x),
            "runtime": cls.place_holder_for_empty(""),
        }

        cls.print_header()
        cls.print_table("Model List", _models, custom_column=custom_column)
        return _models, _pager


class ModelTermViewJson(ModelTermView):
    @classmethod
    def list(  # type: ignore
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.List[str]] = None,
    ) -> None:
        filters = filters or []
        _models, _pager = super().list(
            project_uri, fullname, show_removed, page, size, filters
        )
        cls.pretty_json(_models)

    def info(self, output_filter: ModelInfoFilter = ModelInfoFilter.basic) -> None:
        info = self.model.info()
        if output_filter == ModelInfoFilter.basic:
            info = {"basic": info.get("basic", {})}
        elif output_filter == ModelInfoFilter.model_yaml:
            info = {"model_yaml": info.get("model_yaml", "")}
        elif output_filter == ModelInfoFilter.handlers:
            info = {"handlers": info.get("handlers", {})}
        elif output_filter == ModelInfoFilter.files:
            info = {"files": info.get("files", [])}

        self.pretty_json(info)

    def history(self, fullname: bool = False) -> None:
        fullname = fullname or self.uri.instance.is_cloud
        self.pretty_json(BaseTermView.get_history_data(self.model.history(), fullname))


def get_term_view(ctx_obj: t.Dict) -> t.Type[ModelTermView]:
    return ModelTermViewJson if ctx_obj.get("output") == "json" else ModelTermViewRich



import typing as t
from copy import deepcopy


def flatten(
    origin: t.Dict[str, t.Any], extract_sequence: bool = False
) -> t.Dict[str, t.Any]:
    rt = {}

    def _f_dict(_s: t.Dict[str, t.Any], _prefix: str = "") -> None:
        for _k, _v in _s.items():
            _k = f"{_prefix}{_k}"
            if isinstance(_v, dict):
                _f_dict(_v, _prefix=f"{_k}/")
            elif isinstance(_v, (tuple, list)) and extract_sequence:
                _f_sequence(_v, _prefix=f"{_k}/")
            else:
                rt[_k] = _v

    def _f_sequence(
        data: t.Union[t.Tuple[t.Any, ...], t.List[t.Any]], _prefix: str = ""
    ) -> None:
        index = 0
        for _d in data:
            if isinstance(_d, dict):
                _f_dict(_d, _prefix=f"{_prefix}")
            elif isinstance(_d, (tuple, list)):
                _f_sequence(_d, _prefix=f"{_prefix}")
            else:
                rt[f"{_prefix}{index}"] = _d
            index += 1

    _f_dict(deepcopy(origin))
    return rt


def transform_dict(d: dict, key_selector: dict) -> dict:
    _r = {}
    for field_selector, v in key_selector.items():
        data = d
        fields = field_selector.split(".")
        for field in fields:
            if "[" in field and "]" in field:
                # Parse array index
                index = int(field[field.find("[") + 1 : field.find("]")])
                # Get field name
                field = field[: field.find("[")]
                # Check if field exists and is an array
                if field in data and isinstance(data[field], list):
                    # Check if array index is within bounds
                    if 0 <= index < len(data[field]):
                        data = data[field][index]
                    else:
                        raise ValueError(
                            f"Array index {index} out of bounds for field {field}"
                        )
                else:
                    raise ValueError(f"Field {field} not found or not an array")
            else:
                # Check if field exists
                if field in data:
                    data = data[field]
                else:
                    raise ValueError(f"Field {field} not found in data")
        _r.update({v: data})
    return _r



import os
import abc
import typing as t
import tarfile
import platform
from abc import ABCMeta, abstractmethod
from pathlib import Path
from contextlib import ExitStack

import yaml
from fs.walk import Walker
from typing_extensions import Protocol

from starwhale.utils import console, now_str, gen_uniq_version
from starwhale.consts import (
    YAML_TYPES,
    CREATED_AT_KEY,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    SHORT_VERSION_CNT,
    SW_IGNORE_FILE_NAME,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.version import STARWHALE_VERSION
from starwhale.base.tag import StandaloneTag
from starwhale.utils.fs import move_dir, empty_dir, ensure_dir, ensure_file, extract_tar
from starwhale.base.store import BaseStorage, BundleField
from starwhale.utils.venv import SUPPORTED_PIP_REQ
from starwhale.utils.error import FileTypeError, NotFoundError, MissingFieldError
from starwhale.utils.config import SWCliConfigMixed
from starwhale.base.uri.project import Project
from starwhale.base.uri.resource import Resource


class BaseBundle(metaclass=ABCMeta):
    def __init__(self, uri: Resource) -> None:
        self.uri = uri
        self.name = self.uri.name
        self.sw_config = SWCliConfigMixed()
        self.yaml_name = ""

    @abstractmethod
    def info(self) -> t.Dict[str, t.Any]:
        raise NotImplementedError

    @abstractmethod
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def list_tags(self) -> t.List[str]:
        raise NotImplementedError

    @abstractmethod
    def add_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        raise NotImplementedError

    @abstractmethod
    def remove_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        raise NotImplementedError

    @abstractmethod
    def history(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.List[t.Dict[str, t.Any]]:
        raise NotImplementedError

    @classmethod
    def list(
        cls,
        project_uri: Project,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.Union[t.Dict[str, t.Any], t.List[str]]] = None,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        filters = filters or {}
        _cls = cls._get_cls(project_uri)
        _filter = cls.get_filter_dict(filters, cls.get_filter_fields())
        return _cls.list(project_uri, page, size, _filter)  # type: ignore

    @classmethod
    def get_filter_dict(
        cls,
        filters: t.Union[t.Dict[str, t.Any], t.List[str]],
        fields: t.Optional[t.List[str]] = None,
    ) -> t.Dict[str, t.Any]:
        fields = fields or []
        if isinstance(filters, t.Dict):
            return {k: v for k, v in filters.items() if k in fields}

        _filter_dict: t.Dict[str, t.Any] = {}
        for _f in filters:
            _item = _f.split("=", 1)
            if _item[0] in fields:
                _filter_dict[_item[0]] = _item[1] if len(_item) > 1 else ""
        return _filter_dict

    @classmethod
    def get_filter_fields(cls) -> t.List[str]:
        return ["name", "owner", "latest"]

    @classmethod
    def do_bundle_filter(
        cls,
        bundle_field: BundleField,
        filters: t.Union[t.Dict[str, t.Any], t.List[str]],
    ) -> bool:
        filter_dict = cls.get_filter_dict(filters, cls.get_filter_fields())
        _name = filter_dict.get("name")
        if _name and not bundle_field.name.startswith(_name):
            return False
        _latest = filter_dict.get("latest") is not None
        if _latest and "latest" not in bundle_field.tags:
            return False

        return True

    @classmethod
    @abc.abstractmethod
    def _get_cls(cls, uri: Project) -> t.Any:
        raise NotImplementedError

    @classmethod
    def copy(cls, src_uri: Resource, dest_uri: str) -> None:
        raise NotImplementedError

    def extract(self, force: bool = False, target: t.Union[str, Path] = "") -> Path:
        raise NotImplementedError

    def build(self, *args: t.Any, **kwargs: t.Any) -> None:
        self.store.building = True  # type: ignore

        # use a temp dir to build resources
        # and mv results to dst dir to prevent leaving garbage when interrupt
        def when_exit() -> None:
            src = self.store.snapshot_workdir  # type: ignore
            self.store.building = False  # type: ignore
            dst = self.store.snapshot_workdir  # type: ignore
            ensure_dir(dst.parent)
            os.rename(src, dst)
            console.print(f":100: finish gen resource @ {dst}")

        with ExitStack() as stack:
            stack.callback(when_exit)
            kwargs["yaml_name"] = kwargs.get("yaml_name", self.yaml_name)
            self.buildImpl(*args, **kwargs)

    def buildImpl(self, *args: t.Any, **kwargs: t.Any) -> None:
        raise NotImplementedError

    @property
    def version(self) -> str:
        return getattr(self, "_version", "") or self.uri.version


# https://mypy.readthedocs.io/en/latest/more_types.html#mixin-classes
class LocalStorageBundleProtocol(Protocol):
    uri: Resource
    _version: str
    store: BaseStorage
    tag: StandaloneTag


class LocalStorageBundleMixin(LocalStorageBundleProtocol):
    def __init__(self) -> None:
        self._manifest: t.Dict[str, t.Any] = {}

    def _render_manifest(self) -> None:
        self._manifest["build"] = dict(
            os=platform.system(),
            sw_version=STARWHALE_VERSION,
        )
        # TODO: add signature for import files: model, config
        _fpath = self.store.snapshot_workdir / DEFAULT_MANIFEST_NAME
        ensure_file(_fpath, yaml.safe_dump(self._manifest, default_flow_style=False))

    def _gen_version(self) -> None:
        if not getattr(self, "_version", ""):
            self._version = gen_uniq_version()

        self.uri.version = self._version
        console.debug(f":new: version {self._version[:SHORT_VERSION_CNT]}")
        self._manifest["version"] = self._version
        self._manifest[CREATED_AT_KEY] = now_str()

    def _make_auto_tags(self) -> None:
        self.tag.add_fast_tag()

    def _make_tar(self, ftype: str = "") -> None:
        out = self.store.bundle_dir / f"{self._version}{ftype}"
        ensure_dir(self.store.bundle_dir)

        with tarfile.open(out, "w:") as tar:
            tar.add(str(self.store.snapshot_workdir), arcname="")

        console.print(f":butterfly: {ftype} bundle:{out}")

    @classmethod
    def _do_validate_yaml(cls, path: Path) -> None:
        # add more yaml validators
        if not path.exists():
            raise NotFoundError(str(path))

        if not path.name.endswith(YAML_TYPES):
            raise FileTypeError(f"{path} file type is not yaml|yml")

    def _do_extract(self, force: bool = False, target: t.Union[str, Path] = "") -> Path:
        _store = self.store
        _uri = self.uri

        if not _uri.version:
            raise MissingFieldError("no version")

        _target: Path = Path(target) if target else _store.snapshot_workdir

        if (
            _target.exists()
            and (_target / DEFAULT_MANIFEST_NAME).exists()
            and not force
        ):
            console.print(f":joy_cat: {_target} existed, skip extract model bundle")
        else:
            extract_tar(_store.bundle_path, _target, force)

        return _target

    def _get_src_walker(
        self,
        workdir: Path,
        include_files: t.Optional[t.List[str]] = None,
        exclude_files: t.Optional[t.List[str]] = None,
    ) -> Walker:
        include_files = include_files or []
        exclude_files = exclude_files or []
        _filter = (
            ["*.py", "*.sh", "*.yaml", "*.yml"] + SUPPORTED_PIP_REQ + include_files
        )
        _exclude = exclude_files

        _sw_ignore_path = workdir / SW_IGNORE_FILE_NAME
        if _sw_ignore_path.exists():
            for _l in _sw_ignore_path.read_text().splitlines():
                _l = _l.strip()
                if not _l or _l.startswith("#"):
                    continue

                _exclude.append(_l)

        # Notice: if pass [] as exclude_dirs value, walker will fail
        _exclude = _exclude or None  # type: ignore
        return Walker(filter=_filter, exclude_dirs=_exclude)

    def _do_remove(self, force: bool = False) -> t.Tuple[bool, str]:
        store = self.store

        if force:
            empty_dir(store.loc)
            empty_dir(store.snapshot_workdir)
            return True, ""
        else:
            _ok, _reason = move_dir(store.loc, store.recover_loc, False)
            _ok2, _reason2 = True, ""
            if store.snapshot_workdir.exists():
                _ok2, _reason2 = move_dir(
                    store.snapshot_workdir,
                    store.recover_snapshot_workdir,
                    False,
                )
            return _ok and _ok2, _reason + _reason2
